import pytest
import numpy as np
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LinearRegression, Ridge, LogisticRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.base import is_regressor, is_classifier


from sklego.common import flatten
from sklego.meta import DecayEstimator
from tests.conftest import (
    general_checks,
    classifier_checks,
    regressor_checks,
)


@pytest.mark.parametrize("test_fn", flatten([general_checks, regressor_checks]))
def test_estimator_checks_regression(test_fn):
    trf = DecayEstimator(LinearRegression(), check_input=True)
    test_fn(DecayEstimator.__name__, trf)


@pytest.mark.parametrize("test_fn", flatten([general_checks, classifier_checks]))
def test_estimator_checks_classification(test_fn):
    trf = DecayEstimator(LogisticRegression(solver="lbfgs"), check_input=True)
    test_fn(DecayEstimator.__name__, trf)


@pytest.mark.parametrize(
    "mod", flatten([LinearRegression(), Ridge(), DecisionTreeRegressor()])
)
def test_decay_weight_regr(mod):
    X, y = np.random.normal(0, 1, (100, 100)), np.random.normal(0, 1, (100,))
    mod = DecayEstimator(mod, decay=0.95).fit(X, y)
    assert mod.weights_[0] == pytest.approx(0.95**100, abs=0.001)


@pytest.mark.parametrize(
    "mod", flatten([DecisionTreeClassifier(), LogisticRegression(solver="lbfgs")])
)
def test_decay_weight_clf(mod):
    X, y = (
        np.random.normal(0, 1, (100, 100)),
        (np.random.normal(0, 1, (100,)) < 0).astype(int),
    )
    mod = DecayEstimator(mod, decay=0.95).fit(X, y)
    assert mod.weights_[0] == pytest.approx(0.95**100, abs=0.001)


@pytest.mark.parametrize("mod", flatten([KNeighborsClassifier()]))
def test_throw_warning(mod):
    X, y = np.random.normal(0, 1, (100, 100)), np.random.normal(0, 1, (100,)) < 0
    with pytest.raises(TypeError) as e:
        DecayEstimator(mod, decay=0.95).fit(X, y)
        assert "sample_weight" in str(e)
        assert type(mod).__name__ in str(e)


@pytest.mark.parametrize(
    "mod, is_regr",
    [
        (LinearRegression(), True),
        (Ridge(), True),
        (DecisionTreeRegressor(), True),
        (LogisticRegression(), False),
        (DecisionTreeClassifier(), False),
    ],
)
def test_estimator_type_regressor(mod, is_regr):
    mod = DecayEstimator(mod)
    assert mod._estimator_type == mod.model._estimator_type
    assert is_regressor(mod) == is_regr
    assert is_classifier(mod) == (not is_regr)



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted

from sklego.common import as_list


class OrthogonalTransformer(BaseEstimator, TransformerMixin):
    """
    Transform the columns of a dataframe or numpy array to a column orthogonal or orthonormal matrix.
    Q, R such that X = Q*R, with Q orthogonal, from which follows Q = X*inv(R)
    :param normalize: whether orthogonal matrix should be orthonormal as well
    """

    def __init__(self, normalize=False):
        self.normalize = normalize

    def fit(self, X, y=None):
        """
        Store the inverse of R of the QR decomposition of X, which can be used to calculate the orthogonal projection
        of X. If normalization is required, also stores a vector with normalization terms
        """
        X = check_array(X, estimator=self)

        if not X.shape[0] > 1:
            raise ValueError("Orthogonal transformation not valid for one sample")

        # Q, R such that X = Q*R, with Q orthogonal, from which follows Q = X*inv(R)
        Q, R = np.linalg.qr(X)
        self.inv_R_ = np.linalg.inv(R)

        if self.normalize:
            self.normalization_vector_ = np.linalg.norm(Q, ord=2, axis=0)
        else:
            self.normalization_vector_ = np.ones((X.shape[1],))

        return self

    def transform(self, X):
        """Transforms X using the fitted inverse of R. Normalizes the result if required"""
        if self.normalize:
            check_is_fitted(self, ["inv_R_", "normalization_vector_"])
        else:
            check_is_fitted(self, ["inv_R_"])

        X = check_array(X, estimator=self)

        return X @ self.inv_R_ / self.normalization_vector_


def scalar_projection(vec, unto):
    return vec.dot(unto) / unto.dot(unto)


def vector_projection(vec, unto):
    return scalar_projection(vec, unto) * unto


class InformationFilter(BaseEstimator, TransformerMixin):
    """
    The `InformationFilter` uses a variant of the gram smidt process
    to filter information out of the dataset. This can be useful if you
    want to filter information out of a dataset because of fairness.
    To explain how it works: given a training matrix :math:`X` that contains
    columns :math:`x_1, ..., x_k`. If we assume columns :math:`x_1` and :math:`x_2`
    to be the sensitive columns then the information-filter will
    remove information by applying these transformations;
    .. math::
       \\begin{split}
       v_1 & = x_1 \\\\
       v_2 & = x_2 - \\frac{x_2 v_1}{v_1 v_1}\\\\
       v_3 & = x_3 - \\frac{x_k v_1}{v_1 v_1} - \\frac{x_2 v_2}{v_2 v_2}\\\\
       ... \\\\
       v_k & = x_k - \\frac{x_k v_1}{v_1 v_1} - \\frac{x_2 v_2}{v_2 v_2}
       \\end{split}
    Concatenating our vectors (but removing the sensitive ones) gives us
    a new training matrix :math:`X_{fair} =  [v_3, ..., v_k]`.
    :param columns: the columns to filter out this can be a sequence of either int
                    (in the case of numpy) or string (in the case of pandas).
    :param alpha: parameter to control how much to filter, for alpha=1 we filter out
                  all information while for alpha=0 we don't apply any.
    """

    def __init__(self, columns, alpha=1):
        self.columns = columns
        self.alpha = alpha

    def _check_coltype(self, X):
        for col in as_list(self.columns):
            if isinstance(col, str):
                if isinstance(X, np.ndarray):
                    raise ValueError(
                        f"column {col} is a string but datatype receive is numpy."
                    )
                if isinstance(X, pd.DataFrame):
                    if col not in X.columns:
                        raise ValueError(f"column {col} is not in {X.columns}")
            if isinstance(col, int):
                if col not in range(np.atleast_2d(np.array(X)).shape[1]):
                    raise ValueError(
                        f"column {col} is out of bounds for input shape {X.shape}"
                    )

    def _col_idx(self, X, name):
        if isinstance(name, str):
            if isinstance(X, np.ndarray):
                raise ValueError(
                    "You cannot have a column of type string on a numpy input matrix."
                )
            return {name: i for i, name in enumerate(X.columns)}[name]
        return name

    def _make_v_vectors(self, X, col_ids):
        vs = np.zeros((X.shape[0], len(col_ids)))
        for i, c in enumerate(col_ids):
            vs[:, i] = X[:, col_ids[i]]
            for j in range(0, i):
                vs[:, i] = vs[:, i] - vector_projection(vs[:, i], vs[:, j])
        return vs

    def fit(self, X, y=None):
        """Learn the projection required to make the dataset orthogonal to sensitive columns."""
        self._check_coltype(X)
        self.col_ids_ = [
            v if isinstance(v, int) else self._col_idx(X, v)
            for v in as_list(self.columns)
        ]
        X = check_array(X, estimator=self)
        X_fair = X.copy()
        v_vectors = self._make_v_vectors(X, self.col_ids_)
        # gram smidt process but only on sensitive attributes
        for i, col in enumerate(X_fair.T):
            for v in v_vectors.T:
                X_fair[:, i] = X_fair[:, i] - vector_projection(X_fair[:, i], v)
        # we want to learn matrix P: X P = X_fair
        # this means we first need to create X_fair in order to learn P
        self.projection_, resid, rank, s = np.linalg.lstsq(X, X_fair, rcond=None)
        return self

    def transform(self, X):
        """Transforms X by applying the information filter."""
        check_is_fitted(self, ["projection_", "col_ids_"])
        self._check_coltype(X)
        X = check_array(X, estimator=self)
        # apply the projection and remove the column we won't need
        X_fair = X @ self.projection_
        X_removed = np.delete(X_fair, self.col_ids_, axis=1)
        X_orig = np.delete(X, self.col_ids_, axis=1)
        return self.alpha * np.atleast_2d(X_removed) + (1 - self.alpha) * np.atleast_2d(
            X_orig
        )



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.utils.validation import (
    check_is_fitted,
    check_array
)


class RegressionOutlierDetector(BaseEstimator, OutlierMixin):
    """
    Morphs a regression model into one that can detect outliers. We will try
    to predict `column` in X.
    """
    def __init__(self, model, column, lower=2, upper=2, method='sd'):
        self.model = model
        self.column = column
        self.lower = lower
        self.upper = upper
        self.method = method

    def _is_regression_model(self):
        return any(
            ["RegressorMixin" in p.__name__ for p in type(self.model).__bases__]
        )

    def _handle_thresholds(self, y_true, y_pred):
        difference = y_true - y_pred
        results = np.ones(difference.shape, dtype=int)
        allowed_methods = ["sd", "relative", "absolute"]
        if self.method not in allowed_methods:
            ValueError(f"`method` must be in {allowed_methods} got: {self.method}")
        if self.method == "sd":
            lower_limit_hit = -self.lower * self.sd_ > difference
            upper_limit_hit = self.upper * self.sd_ < difference
        if self.method == "relative":
            lower_limit_hit = -self.lower > difference/y_true
            upper_limit_hit = self.upper < difference/y_true
        if self.method == "absolute":
            lower_limit_hit = -self.lower > difference
            upper_limit_hit = self.upper < difference
        results[lower_limit_hit] = -1
        results[upper_limit_hit] = -1
        return results

    def to_x_y(self, X):
        y = X[:, self.idx_]
        cols_to_use = [i for i in range(X.shape[1]) if i != self.column]
        X_to_use = X[:, cols_to_use]
        if len(X_to_use.shape) == 1:
            X_to_use = X_to_use.reshape(-1, 1)
        return X_to_use, y

    def fit(self, X, y=None):
        """
        Fit the data after adapting the same weight.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :param y: array-like, shape=(n_samples,) training data.
        :return: Returns an instance of self.
        """
        self.idx_ = np.argmax([i == self.column for i in X.columns]) if isinstance(X, pd.DataFrame) else self.column
        X = check_array(X, estimator=self)
        if not self._is_regression_model():
            raise ValueError("Passed model must be regression!")
        X, y = self.to_x_y(X)
        self.estimator_ = self.model.fit(X, y)
        self.sd_ = np.std(self.estimator_.predict(X) - y)
        return self

    def predict(self, X, y=None):
        """
        Predict new data.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :return: array, shape=(n_samples,) the predicted data
        """
        check_is_fitted(self, ['estimator_', 'sd_', 'idx_'])
        X = check_array(X, estimator=self)
        X, y = self.to_x_y(X)
        preds = self.estimator_.predict(X)
        return self._handle_thresholds(y, preds)

    def score_samples(self, X, y=None):
        check_is_fitted(self, ['estimator_', 'sd_', 'idx_'])
        X = check_array(X, estimator=self)
        X, y_true = self.to_x_y(X)
        y_pred = self.estimator_.predict(X)
        difference = y_true - y_pred
        allowed_methods = ["sd", "relative", "absolute"]
        if self.method not in allowed_methods:
            ValueError(f"`method` must be in {allowed_methods} got: {self.method}")
        if self.method == "sd":
            return difference
        if self.method == "relative":
            return difference/y_true
        if self.method == "absolute":
            return difference



"""Test the LADRegressor."""

import numpy as np
import pytest

from sklego.linear_model import LADRegression
from sklego.testing import check_shape_remains_same_regressor
from sklego.common import flatten
from tests.conftest import general_checks, nonmeta_checks, select_tests, regressor_checks

test_batch = [
    (np.array([0, 0, 3, 0, 6]), 3),
    (np.array([1, 0, -2, 0, 4, 0, -5, 0, 6]), 2),
    (np.array([4, -4]), 0),
]


def _create_dataset(coefs, intercept, noise=0.0):
    np.random.seed(0)
    X = np.random.randn(10000, coefs.shape[0])
    y = X @ coefs + intercept + noise * np.random.randn(10000)

    return X, y


@pytest.mark.parametrize("method", ["SLSQP", "TNC", "L-BFGS-B"])
@pytest.mark.parametrize("coefs, intercept", test_batch)
def test_coefs_and_intercept__no_noise(coefs, intercept, method):
    """Regression problems without noise."""
    X, y = _create_dataset(coefs, intercept)
    lad = LADRegression(method=method)
    lad.fit(X, y)
    assert lad.score(X, y) > 0.99


@pytest.mark.parametrize("method", ["SLSQP", "TNC", "L-BFGS-B"])
@pytest.mark.parametrize("coefs, intercept", test_batch)
def test_score(coefs, intercept, method):
    """Tests with noise on an easy problem. A good score should be possible."""
    X, y = _create_dataset(coefs, intercept, noise=1.0)
    lad = LADRegression(method=method)
    lad.fit(X, y)
    assert lad.score(X, y) > 0.9


@pytest.mark.parametrize("method", ["SLSQP", "TNC", "L-BFGS-B"])
@pytest.mark.parametrize("coefs, intercept", test_batch)
def test_coefs_and_intercept__no_noise_positive(coefs, intercept, method):
    """Test with only positive coefficients."""
    X, y = _create_dataset(coefs, intercept, noise=0.0)
    lad = LADRegression(method=method, positive=True)
    lad.fit(X, y)
    assert all(lad.coef_ >= 0)
    assert lad.score(X, y) > 0.3


@pytest.mark.parametrize("coefs, intercept", test_batch)
def test_coefs_and_intercept__no_noise_regularization(coefs, intercept):
    """Test model with regularization. The size of the coef vector should shrink the larger alpha gets."""
    X, y = _create_dataset(coefs, intercept)

    lads = [LADRegression(alpha=alpha, l1_ratio=0.).fit(X, y) for alpha in range(3)]
    coef_size = np.array([np.sum(lad.coef_ ** 2) for lad in lads])

    for i in range(2):
        assert coef_size[i] >= coef_size[i + 1]


@pytest.mark.parametrize("coefs, intercept", test_batch)
def test_fit_intercept_and_copy(coefs, intercept):
    """Test if fit_intercept and copy_X work."""
    X, y = _create_dataset(coefs, intercept, noise=2.0)
    imb = LADRegression(fit_intercept=False, copy_X=False)
    imb.fit(X, y)

    assert imb.intercept_ == 0.0


@pytest.mark.parametrize("test_fn", [check_shape_remains_same_regressor])
def test_lad(test_fn):
    regr = LADRegression()
    test_fn(LADRegression.__name__, regr)


@pytest.mark.parametrize("positive", [True, False])
@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize("method", ["SLSQP", "TNC", "L-BFGS-B"])
@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, regressor_checks]),
    )
)
def test_estimator_checks(positive, fit_intercept, method, test_fn):
    regr = (
        f'{LADRegression.__name__}_method_{method}_positive_{positive}_fit_intercept_{fit_intercept}',
        LADRegression(method=method, positive=positive, fit_intercept=fit_intercept)
    )
    test_fn(*regr)



import pytest
from sklego.common import as_list, sliding_window


def test_as_list_strings():
    assert as_list("test") == ["test"]
    assert as_list(["test1", "test2"]) == ["test1", "test2"]


def test_as_list_ints():
    assert as_list(123) == [123]
    assert as_list([1, 2, 3]) == [1, 2, 3]


def test_as_list_other():
    def f():
        return 123

    assert as_list(f) == [f]
    assert as_list(range(1, 4)) == [1, 2, 3]


@pytest.mark.parametrize(
    "sequence, window_size, step_size",
    [([1, 2, 3, 4, 5], 2, 1), (["a", "b", "c", "d", "e"], 3, 2)],
)
def test_sliding_window(sequence, window_size, step_size):
    windows = list(sliding_window(sequence, window_size, step_size))
    assert windows[0] == sequence[:window_size]
    assert len(windows[0]) == window_size
    assert windows[1][0] == sequence[step_size]



__all__ = [
    "ConfusionBalancer",
    "DecayEstimator",
    "EstimatorTransformer",
    "GroupedEstimator",
    "GroupedPredictor",
    "GroupedTransformer",
    "OutlierRemover",
    "SubjectiveClassifier",
    "Thresholder",
    "RegressionOutlierDetector",
    "OutlierClassifier",
    "ZeroInflatedRegressor"
]

from .confusion_balancer import ConfusionBalancer
from .decay_estimator import DecayEstimator
from .estimator_transformer import EstimatorTransformer
from .grouped_estimator import GroupedEstimator
from .grouped_predictor import GroupedPredictor
from .grouped_transformer import GroupedTransformer
from .outlier_remover import OutlierRemover
from .subjective_classifier import SubjectiveClassifier
from .thresholder import Thresholder
from .regression_outlier_detector import RegressionOutlierDetector
from .outlier_classifier import OutlierClassifier
from .zero_inflated_regressor import ZeroInflatedRegressor



import pytest
import numpy as np
from unittest.mock import Mock
from unittest.mock import patch
from sklearn import clone
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import HistGradientBoostingClassifier
from sklearn.multioutput import MultiOutputRegressor
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.preprocessing import OrdinalEncoder
from sklearn.utils import check_X_y

from sklego.common import flatten
from sklego.meta import EstimatorTransformer
from tests.conftest import transformer_checks, general_checks


@pytest.mark.parametrize(
    "test_fn", flatten([transformer_checks, general_checks])
)
def test_estimator_checks(test_fn):
    trf = EstimatorTransformer(LinearRegression(), check_input=True)
    test_fn(EstimatorTransformer.__name__, trf)


def test_values_uniform(random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    X, y = check_X_y(X, y)
    clf = DummyClassifier(strategy="most_frequent")
    transformer = EstimatorTransformer(clone(clf))
    transformed = transformer.fit(X, y).transform(X)

    assert transformed.shape == (y.shape[0], 1)
    assert np.all(transformed == clf.fit(X, y).predict(X))


def test_set_params():
    clf = DummyClassifier(strategy="most_frequent")
    transformer = EstimatorTransformer(clf)

    transformer.set_params(estimator__strategy="stratified")
    assert clf.strategy == "stratified"


def test_get_params():
    clf = DummyClassifier(strategy="most_frequent")
    transformer = EstimatorTransformer(clf)

    assert transformer.get_params() == {
        "estimator__constant": None,
        "estimator__random_state": None,
        "estimator": clf,
        "estimator__strategy": "most_frequent",
        "predict_func": "predict",
        "check_input": False,
    }


def test_shape(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    m = X.shape[0]
    pipeline = Pipeline(
        [
            (
                "ml_features",
                FeatureUnion(
                    [
                        ("model_1", EstimatorTransformer(LinearRegression())),
                        ("model_2", EstimatorTransformer(Ridge())),
                    ]
                ),
            )
        ]
    )

    assert pipeline.fit(X, y).transform(X).shape == (m, 2)


def test_shape_multitarget(random_xy_dataset_multitarget):
    X, y = random_xy_dataset_multitarget
    m = X.shape[0]
    n = y.shape[1]
    pipeline = Pipeline(
        [
            (
                "multi_ml_features",
                EstimatorTransformer(MultiOutputRegressor(Ridge()))
            )
        ]
    )
    assert pipeline.fit(X, y).transform(X).shape == (m, n)


@patch('sklego.meta.estimator_transformer.clone')
def test_kwargs(patched_clone, random_xy_dataset_clf):
    """ Test if kwargs are properly passed to an underlying estimator. """
    X, y = random_xy_dataset_clf
    estimator = Mock()
    patched_clone.return_value = estimator

    sample_weights = np.ones(shape=len(y))
    pipeline = EstimatorTransformer(estimator)
    pipeline.fit(X, y, sample_weight=sample_weights)

    # We can't use `assert_called_with` because that compares by `==` which is ambiguous
    # on numpy arrays
    np.testing.assert_array_equal(
        X, estimator.fit.call_args[0][0]
    )
    np.testing.assert_array_equal(
        y, estimator.fit.call_args[0][1]
    )
    np.testing.assert_array_equal(
        sample_weights, estimator.fit.call_args[1]['sample_weight']
    )


@pytest.mark.parametrize(
    "X", [np.array([[np.nan, 4], [7, 3], [5, 5], [7, 2], [5, 7]]),
          np.array([["a", 4], ["a", 3], ["b", 5], ["b", 2], ["b", 7]])]
)
def test_nan_and_string_input(X):
    """ Test X containing nan and string with check_input=False. """
    y = np.array([1, 0, 1, 0, 1])
    clf = Pipeline(
        [
            ("encoder", OrdinalEncoder()),
            ("gbm", HistGradientBoostingClassifier())
        ]
    )
    transformer = EstimatorTransformer(clf, check_input=False)
    transformed = transformer.fit(X, y).transform(X)

    assert transformed.shape == (y.shape[0], 1)
    assert np.all(transformed == clf.fit(X, y).predict(X))



import pytest
import numpy as np
from sklearn.ensemble import HistGradientBoostingClassifier
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import StackingClassifier
from sklego.common import flatten
from sklego.meta import Thresholder
from tests.conftest import general_checks, classifier_checks, select_tests


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, classifier_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_fit2d_predict1d",
            "check_methods_subset_invariance",
            "check_dont_overwrite_parameters",
            "check_classifiers_classes",
            "check_classifiers_train",
            "check_supervised_y_2d",
            "check_classifier_data_not_an_array", # https://github.com/koaning/scikit-lego/issues/490
        ]
        # outliers train wont work because we have two thresholds
    ),
)
def test_standard_checks(test_fn):
    trf = Thresholder(LogisticRegression(), threshold=0.5)
    test_fn(Thresholder.__name__, trf)


def test_same_threshold():
    mod1 = Thresholder(LogisticRegression(), threshold=0.5)
    mod2 = LogisticRegression()
    X = np.random.normal(0, 1, (100, 3))
    y = np.random.normal(0, 1, (100,)) < 0
    assert (mod1.fit(X, y).predict(X) == mod2.fit(X, y).predict(X)).all()


def test_diff_threshold():
    mod1 = Thresholder(LogisticRegression(), threshold=0.5)
    mod2 = Thresholder(LogisticRegression(), threshold=0.7)
    mod3 = Thresholder(LogisticRegression(), threshold=0.9)
    np.random.seed(42)
    X = np.random.normal(0, 1, (100, 3))
    y = np.random.normal(0, 1, (100,)) < 0
    assert mod1.fit(X, y).predict(X).sum() >= mod2.fit(X, y).predict(X).sum()
    assert mod2.fit(X, y).predict(X).sum() >= mod3.fit(X, y).predict(X).sum()


def test_raise_error1():
    with pytest.raises(ValueError):
        # we only support classification models
        mod = Thresholder(LinearRegression(), threshold=0.7)
        np.random.seed(42)
        X = np.random.normal(0, 1, (100, 3))
        y = np.random.normal(0, 1, (100,)) < 0
        mod.fit(X, y)


def test_raise_error2():
    with pytest.raises(ValueError):
        mod = Thresholder(LinearRegression(), threshold=0.7)
        np.random.seed(42)
        X = np.random.normal(0, 1, (1000, 3))
        # we only support two classes
        y = np.random.choice(["a", "b", "c"], 1000)
        mod.fit(X, y)


def test_refit_always():
    lr = LogisticRegression()
    mod = Thresholder(lr, threshold=0.5, refit=True)
    np.random.seed(42)
    X = np.random.normal(0, 1, (100, 3))
    y = np.random.normal(0, 1, (100,)) < 0
    assert mod.fit(X, y).predict(X).shape == y.shape


def test_refit_auto():
    lr = LogisticRegression()
    mod = Thresholder(lr, threshold=0.5, refit=False)
    np.random.seed(42)
    X = np.random.normal(0, 1, (100, 3))
    y = np.random.normal(0, 1, (100,)) < 0
    mod.fit(X, y).predict(X)
    assert mod.fit(X, y).predict(X).shape == y.shape


@pytest.mark.parametrize('refit', [True, False])
def test_passes_sample_weight(refit):
    class CustomLR(LogisticRegression):
        def fit(self, X, y, sample_weight=None):
            assert sample_weight is not None
            super().fit(X, y)

    mod = Thresholder(CustomLR(), threshold=0.5, refit=refit)
    np.random.seed(42)
    X = np.random.normal(0, 1, (100, 3))
    y = np.random.normal(0, 1, (100,)) < 0
    weight = np.random.random(100)

    mod.fit(X, y, sample_weight=weight)


def test_no_refit_does_not_fit_underlying():
    X = np.array([1, 2, 3, 4]).reshape(-1, 1)
    y_ones = np.array([0, 1, 1, 1]).reshape(-1, )
    y_zeros = np.array([0, 0, 0, 1]).reshape(-1, )

    clf = DummyClassifier(strategy="most_frequent")
    clf.fit(X, y_ones)
    a = Thresholder(clf, threshold=0.2, refit=False)      
    a.fit(X, y_zeros)

    assert a.predict(np.array([[1]])) == 1


def test_refit_fits_underlying():
    X = np.array([1, 2, 3, 4]).reshape(-1, 1)
    y_ones = np.array([0, 1, 1, 1]).reshape(-1, )
    y_zeros = np.array([0, 0, 0, 1]).reshape(-1, )

    clf = DummyClassifier(strategy="most_frequent")
    clf.fit(X, y_ones)
    a = Thresholder(clf, threshold=0.2, refit=True)      
    a.fit(X, y_zeros)

    assert a.predict(np.array([[1]])) == 0


def test_stacking_classifier():
    '''
    Tests issue https://github.com/koaning/scikit-lego/issues/501

    No asserts are added as we only test for being exception free.
    When cloning the model in Thresholder an unfitted model is generated
    where no predict_proba exists
    '''
    estimators = [("dummy", DummyClassifier(strategy="constant", constant=0))]

    X = np.random.normal(0, 1, (100, 3))
    y = np.random.normal(0, 1, (100,)) < 0

    clf = StackingClassifier(estimators=estimators, final_estimator=DummyClassifier(strategy="constant", constant=0))

    clf.fit(X, y)

    a = Thresholder(clf, threshold=0.2)
    a.fit(X, y)
    a.predict(X)


def test_nans_could_work():
    X = np.array([[np.nan, 4], [7, 3], [5, 5], [7, 2], [5, 7]])
    y = np.array([1, 0, 1, 0, 1])
    model = Thresholder(HistGradientBoostingClassifier(), 0.6)
    model.fit(X, y)



from typing import Tuple

import numpy as np
import pandas as pd
from scipy.sparse import issparse
from sklearn.utils import check_array
from sklearn.utils.validation import _ensure_no_complex_data

from sklego.common import as_list


def constant_shrinkage(group_sizes: list, alpha: float) -> np.ndarray:
    r"""
    The augmented prediction for each level is the weighted average between its prediction and the augmented
    prediction for its parent.

    Let $\hat{y}_i$ be the prediction at level $i$, with $i=0$ being the root, than the augmented prediction
    $\hat{y}_i^* = \alpha \hat{y}_i + (1 - \alpha) \hat{y}_{i-1}^*$, with $\hat{y}_0^* = \hat{y}_0$.
    """
    return np.array(
        [alpha ** (len(group_sizes) - 1)]
        + [
            alpha ** (len(group_sizes) - 1 - i) * (1 - alpha)
            for i in range(1, len(group_sizes) - 1)
        ]
        + [(1 - alpha)]
    )


def relative_shrinkage(group_sizes: list) -> np.ndarray:
    """Weigh each group according to it's size"""
    return np.array(group_sizes)


def min_n_obs_shrinkage(group_sizes: list, min_n_obs) -> np.ndarray:
    """Use only the smallest group with a certain amount of observations"""
    if min_n_obs > max(group_sizes):
        raise ValueError(
            f"There is no group with size greater than or equal to {min_n_obs}"
        )

    res = np.zeros(len(group_sizes))
    res[np.argmin(np.array(group_sizes) >= min_n_obs) - 1] = 1
    return res


def _split_groups_and_values(
    X, groups, name="", min_value_cols=1, check_X=True, **kwargs
) -> Tuple[pd.DataFrame, np.ndarray]:
    _data_format_checks(X, name=name)
    _shape_check(X, min_value_cols)

    try:
        if isinstance(X, pd.DataFrame):
            X_group = X.loc[:, as_list(groups)]
            X_value = X.drop(columns=groups).values
        else:
            X_group = pd.DataFrame(X[:, as_list(groups)])
            pos_indexes = range(X.shape[1])
            X_value = np.delete(X, [pos_indexes[g] for g in as_list(groups)], axis=1)
    except (KeyError, IndexError):
        raise ValueError(f"Could not drop groups {groups} from columns of X")

    X_group = _check_grouping_columns(X_group, **kwargs)

    if check_X:
        X_value = check_array(X_value, **kwargs)

    return X_group, X_value


def _data_format_checks(X, name):
    _ensure_no_complex_data(X)

    if issparse(X):  # sklearn.validation._ensure_sparse_format to complicated
        raise ValueError(f"The estimator {name} does not work on sparse matrices")


def _shape_check(X, min_value_cols):
    if min_value_cols > 1:
        if X.ndim == 1 or X.shape[1] < 2:
            raise ValueError(
                f"0 feature(s) (shape={X.shape}) while a minimum of {min_value_cols} is required."
            )
    else:
        if X.ndim == 2 and X.shape[1] < 1:
            raise ValueError(
                f"0 feature(s) (shape={X.shape}) while a minimum of {min_value_cols} is required."
            )


def _check_grouping_columns(X_group, **kwargs) -> pd.DataFrame:
    """Do basic checks on grouping columns"""
    # Do regular checks on numeric columns
    X_group_num = X_group.select_dtypes(include="number")
    if X_group_num.shape[1]:
        check_array(X_group_num, **kwargs)

    # Only check missingness in object columns
    if X_group.select_dtypes(exclude="number").isnull().any(axis=None):
        raise ValueError("X has NaN values")

    # The grouping part we always want as a DataFrame with range index
    return X_group.reset_index(drop=True)



import pytest
import pandas as pd
import itertools as it
import numpy as np
from sklego.preprocessing import PandasTypeSelector
from tests.conftest import id_func


@pytest.mark.parametrize(
    "transformer", [PandasTypeSelector(include=["number"])], ids=id_func
)
def test_len_regression(transformer, random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    X = pd.DataFrame(X)
    assert transformer.fit(X, y).transform(X).shape[0] == X.shape[0]


@pytest.mark.parametrize(
    "transformer", [PandasTypeSelector(include=["number"])], ids=id_func
)
def test_len_classification(transformer, random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    X = pd.DataFrame(X)
    assert transformer.fit(X, y).transform(X).shape[0] == X.shape[0]


@pytest.mark.parametrize(
    "include,exclude",
    [
        _
        for _ in it.combinations(
            ["number", "datetime", "timedelta", "category", "datetimetz", None], 2
        )
    ],
)
def test_get_params_str(include, exclude):
    transformer = PandasTypeSelector(include=include, exclude=exclude)

    assert transformer.get_params() == {"include": include, "exclude": exclude}


@pytest.mark.parametrize(
    "include,exclude",
    [
        _
        for _ in it.combinations(
            [np.int64, np.float64, np.datetime64, np.timedelta64], 2
        )
    ],
)
def test_get_params_np(include, exclude):
    transformer = PandasTypeSelector(include=include, exclude=exclude)

    assert transformer.get_params() == {"include": include, "exclude": exclude}


def test_value_error_differrent_dtyes():
    fit_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    transform_df = pd.DataFrame({"a": [4, 5, 6], "b": ["4", "5", "6"]})
    transformer = PandasTypeSelector(exclude=["category"]).fit(fit_df)

    with pytest.raises(ValueError):
        transformer.transform(transform_df)


def test_get_feature_names():
    df = pd.DataFrame({"a": [4, 5, 6], "b": ["4", "5", "6"]})
    transformer_number = PandasTypeSelector(include="number").fit(df)
    assert transformer_number.get_feature_names() == ["a"]

    transformer_number = PandasTypeSelector(include="object").fit(df)
    assert transformer_number.get_feature_names() == ["b"]


def test_value_error_empty(random_xy_dataset_regr):
    transformer = PandasTypeSelector(exclude=["number"])
    X, y = random_xy_dataset_regr
    X = pd.DataFrame(X)

    with pytest.raises(ValueError):
        transformer.fit(X, y)


def test_value_error_inequal(random_xy_dataset_regr):
    transformer = PandasTypeSelector(include=["number"])
    X, y = random_xy_dataset_regr
    X = pd.DataFrame(X)
    if X.shape[0] > 0:
        with pytest.raises(ValueError):
            transformer.fit(X)
            # Remove column to create error
            transformer.transform(X.iloc[:, :-1])



import pytest
import numpy as np
import pandas as pd

from sklearn.utils.validation import FLOAT_DTYPES
from sklego.common import flatten
from sklego.preprocessing import ColumnCapper
from tests.conftest import select_tests, transformer_checks, general_checks, nonmeta_checks


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, transformer_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_estimators_nan_inf",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
def test_estimator_checks(test_fn):
    test_fn(ColumnCapper.__name__, ColumnCapper())


def test_quantile_range():
    def expect_type_error(quantile_range):
        with pytest.raises(TypeError):
            ColumnCapper(quantile_range)

    def expect_value_error(quantile_range):
        with pytest.raises(ValueError):
            ColumnCapper(quantile_range)

    # Testing quantile_range type
    expect_type_error(quantile_range=1)
    expect_type_error(quantile_range="a")
    expect_type_error(quantile_range={})
    expect_type_error(quantile_range=set())

    # Testing quantile_range values
    # Invalid type:
    expect_type_error(quantile_range=("a", 90))
    expect_type_error(quantile_range=(10, "a"))

    # Invalid limits
    expect_value_error(quantile_range=(-1, 90))
    expect_value_error(quantile_range=(10, 110))

    # Invalid order
    expect_value_error(quantile_range=(60, 40))


def test_interpolation():
    valid_interpolations = ("linear", "lower", "higher", "midpoint", "nearest")
    invalid_interpolations = ("test", 42, None, [], {}, set(), 0.42)

    for interpolation in valid_interpolations:
        ColumnCapper(interpolation=interpolation)

    for interpolation in invalid_interpolations:
        with pytest.raises(ValueError):
            ColumnCapper(interpolation=interpolation)


@pytest.fixture()
def valid_df():
    return pd.DataFrame(
        {"a": [1, np.nan, 3, 4], "b": [11, 12, np.inf, 14], "c": [21, 22, 23, 24]}
    )


def test_X_types_and_transformed_shapes(valid_df):
    def expect_value_error(X, X_transform=None):
        if X_transform is None:
            X_transform = X
        with pytest.raises(ValueError):
            capper = ColumnCapper().fit(X)
            capper.transform(X_transform)

    # Fitted and transformed arrays must have the same number of columns
    expect_value_error(valid_df, valid_df[["a", "b"]])

    invalid_dfs = [
        pd.DataFrame({"a": [np.nan, np.nan, np.nan], "b": [11, 12, 13]}),
        pd.DataFrame({"a": [np.inf, np.inf, np.inf], "b": [11, 12, 13]}),
    ]

    for invalid_df in invalid_dfs:
        expect_value_error(invalid_df)  # contains an invalid column ('a')
        expect_value_error(
            invalid_df["b"]
        )  # 1d arrays should be reshaped before fitted/transformed
        # Like this:
        ColumnCapper().fit_transform(invalid_df["b"].values.reshape(-1, 1))
        ColumnCapper().fit_transform(invalid_df["b"].values.reshape(1, -1))

    capper = ColumnCapper()
    for X in valid_df, valid_df.values:
        assert capper.fit_transform(X).shape == X.shape


def test_nan_inf(valid_df):
    # Capping infs
    capper = ColumnCapper(discard_infs=False)
    assert (capper.fit_transform(valid_df) == np.inf).sum().sum() == 0
    assert np.isnan(capper.fit_transform(valid_df)).sum() == 1

    # Discarding infs
    capper = ColumnCapper(discard_infs=True)
    assert (capper.fit_transform(valid_df) == np.inf).sum().sum() == 0
    assert np.isnan(capper.fit_transform(valid_df)).sum() == 2


def test_dtype_regression(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    assert ColumnCapper().fit(X, y).transform(X).dtype in FLOAT_DTYPES


def test_dtype_classification(random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    assert ColumnCapper().fit(X, y).transform(X).dtype in FLOAT_DTYPES



import numpy as np
import pytest

from sklego.common import flatten
from sklego.linear_model import ProbWeightRegression
from tests.conftest import nonmeta_checks, regressor_checks, general_checks, select_tests


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, regressor_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
@pytest.mark.cvxpy
def test_estimator_checks(test_fn):
    regr_min_zero = ProbWeightRegression(non_negative=True)
    test_fn(ProbWeightRegression.__name__ + "_min_zero_true", regr_min_zero)
    regr_not_min_zero = ProbWeightRegression(non_negative=False)
    test_fn(ProbWeightRegression.__name__ + "_min_zero_true_false", regr_not_min_zero)


@pytest.mark.cvxpy
def test_shape_trained_model(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    mod_no_intercept = ProbWeightRegression()
    assert mod_no_intercept.fit(X, y).coefs_.shape == (X.shape[1],)
    np.testing.assert_approx_equal(
        mod_no_intercept.fit(X, y).coefs_.sum(), 1.0, significant=4
    )



from sklearn.base import BaseEstimator
from sklearn.utils import check_array, check_X_y
from sklearn.utils.validation import FLOAT_DTYPES, check_random_state, check_is_fitted

from sklego.common import TrainOnlyTransformerMixin


class RandomAdder(TrainOnlyTransformerMixin, BaseEstimator):
    def __init__(self, noise=1, random_state=None):
        self.noise = noise
        self.random_state = random_state

    def fit(self, X, y):
        super().fit(X, y)
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)
        self.dim_ = X.shape[1]

        return self

    def transform_train(self, X):
        rs = check_random_state(self.random_state)
        check_is_fitted(self, ["dim_"])

        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)

        return X + rs.normal(0, self.noise, size=X.shape)



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_X_y
from sklearn.utils.validation import (
    check_is_fitted,
    check_array,
    check_random_state,
    FLOAT_DTYPES,
)


class RandomRegressor(BaseEstimator, RegressorMixin):
    """
    A RandomRegressor makes random predictions only based on the "y"
    value that is seen. The goal is that such a regressor can be used
    for benchmarking. It should be easily beatable.

    :param str strategy: how we want to select random values, can be "uniform" or "normal"
    :param int seed: the seed value, default: 42
    """

    def __init__(self, strategy="uniform", random_state=None):
        self.allowed_strategies = ("uniform", "normal")
        self.random_state = random_state
        self.strategy = strategy

    def fit(self, X: np.array, y: np.array) -> "RandomRegressor":
        """
        Fit the model using X, y as training data.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :param y: array-like, shape=(n_samples,) training data.
        :return: Returns an instance of self.
        """
        if self.strategy not in self.allowed_strategies:
            raise ValueError(
                f"strategy {self.strategy} is not in {self.allowed_strategies}"
            )
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)
        self.dim_ = X.shape[1]

        self.min_ = np.min(y)
        self.max_ = np.max(y)
        self.mu_ = np.mean(y)
        self.sigma_ = np.std(y)

        return self

    def predict(self, X):
        """
        Predict new data by making random guesses.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :return: array, shape=(n_samples,) the predicted data
        """
        rs = check_random_state(self.random_state)
        check_is_fitted(self, ["dim_", "min_", "max_", "mu_", "sigma_"])

        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        if X.shape[1] != self.dim_:
            raise ValueError(
                f"Unexpected input dimension {X.shape[1]}, expected {self.dim_}"
            )

        if self.strategy == "normal":
            return rs.normal(self.mu_, self.sigma_, X.shape[0])
        if self.strategy == "uniform":
            return rs.uniform(self.min_, self.max_, X.shape[0])



import numpy as np
from sklearn.decomposition import PCA
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.utils.validation import check_is_fitted, check_array, FLOAT_DTYPES


class PCAOutlierDetection(BaseEstimator, OutlierMixin):
    """
    Does outlier detection based on the reconstruction error from PCA.
    """

    def __init__(
        self,
        n_components=None,
        threshold=None,
        variant="relative",
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
    ):
        self.n_components = n_components
        self.threshold = threshold
        self.whiten = whiten
        self.variant = variant
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state

    def fit(self, X, y=None):
        """
        Fit the model using X as training data.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :param y: ignored but kept in for pipeline support
        :return: Returns an instance of self.
        """
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        if not self.threshold:
            raise ValueError("The `threshold` value cannot be `None`.")

        self.pca_ = PCA(
            n_components=self.n_components,
            whiten=self.whiten,
            svd_solver=self.svd_solver,
            tol=self.tol,
            iterated_power=self.iterated_power,
            random_state=self.random_state,
        )
        self.pca_.fit(X, y)
        self.offset_ = -self.threshold
        return self

    def transform(self, X):
        """
        Uses the underlying PCA method to transform the data.
        """
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        check_is_fitted(self, ["pca_", "offset_"])
        return self.pca_.transform(X)

    def difference(self, X):
        """
        Shows the calculated difference between original and reconstructed data. Row by row.

        :param X: array-like, shape=(n_columns, n_samples, ) training data.
        :return: array, shape=(n_samples,) the difference
        """
        check_is_fitted(self, ["pca_", "offset_"])
        reduced = self.pca_.transform(X)
        diff = np.sum(np.abs(self.pca_.inverse_transform(reduced) - X), axis=1)
        if self.variant == "relative":
            diff = diff / X.sum(axis=1)
        return diff

    def decision_function(self, X):
        return self.threshold - self.difference(X)

    def score_samples(self, X):
        return -self.difference(X)

    def predict(self, X):
        """
        Predict if a point is an outlier.

        :param X: array-like, shape=(n_columns, n_samples, ) training data.
        :return: array, shape=(n_samples,) the predicted data. 1 for inliers, -1 for outliers.
        """
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        check_is_fitted(self, ["pca_", "offset_"])
        result = np.ones(X.shape[0])
        result[self.difference(X) > self.threshold] = -1
        return result.astype(int)



try:
    import cvxpy as cp
except ImportError:
    from sklego.notinstalled import NotInstalledPackage

    cp = NotInstalledPackage("cvxpy")

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array, check_X_y
from sklearn.utils.validation import check_is_fitted


def _mk_monotonic_average(xs, ys, intervals, method="increasing", **kwargs):
    """
    Creates smoothed averages of `ys` at the intervals given by `intervals`.
    :param xs: all the datapoints of a feature (represents the x-axis)
    :param ys: all the datapoints what we'd like to predict (represents the y-axis)
    :param intervals: the intervals at which we'd like to get a good average value
    :param method: the method that is used for smoothing, can be either `increasing` or `decreasing`.
    :return:
        An array as long as `intervals` that represents the average `y`-values at those intervals,
        keeping the constraint in mind.
    """
    x_internal = np.array([xs >= i for i in intervals]).T.astype(float)
    betas = cp.Variable(x_internal.shape[1])
    objective = cp.Minimize(cp.sum_squares(x_internal @ betas - ys))
    if method == "increasing":
        constraints = [betas[i + 1] >= 0 for i in range(betas.shape[0] - 1)]
    elif method == "decreasing":
        constraints = [betas[i + 1] <= 0 for i in range(betas.shape[0] - 1)]
    else:
        raise ValueError(
            f"method must be either `increasing` or `decreasing`, got: {method}"
        )
    prob = cp.Problem(objective, constraints)
    prob.solve()
    return betas.value.cumsum()


def _mk_average(xs, ys, intervals, method="average", span=1, **kwargs):
    """
    Creates smoothed averages of `ys` at the intervals given by `intervals`.
    :param xs: all the datapoints of a feature (represents the x-axis)
    :param ys: all the datapoints what we'd like to predict (represents the y-axis)
    :param intervals: the intervals at which we'd like to get a good average value
    :param method: the method that is used for smoothing, can be either `average` or `normal`.
    :param span: if the method is `average` then this is the span around the interval
    that is used to determine the average `y`-value, if the method is `normal` the span
    becomes the value of sigma that is used for weighted averaging
    :return:
        An array as long as `intervals` that represents the average `y`-values at those intervals.
    """
    results = np.zeros(intervals.shape)
    for idx, interval in enumerate(intervals):
        if method == "average":
            distances = 1 / (0.01 + np.abs(xs - interval))
            predicate = (xs < (interval + span)) | (xs < (interval - span))
        elif method == "normal":
            distances = np.exp(-((xs - interval) ** 2) / span)
            predicate = xs == xs
        else:
            raise ValueError("method needs to be either `average` or `normal`")
        subset = ys[predicate]
        dist_subset = distances[predicate]
        results[idx] = np.average(subset, weights=dist_subset)
    return results


class IntervalEncoder(TransformerMixin, BaseEstimator):
    """
    The interval encoder bends features in `X` with regards to`y`.
    We take each column in X separately and smooth it towards `y` using
    the strategy that is defined in `method`.
    Note that this allows us to make certain features strictly monotonic
    in your machine learning model if you follow this with an appropriate
    model.
    :param n_chunks: the number of cuts that makes the interval
    :param method: the interpolation method used, must be in
    ["average", "normal", "increasing", "decreasing"], default: "normal"
    :param span: a hyperparameter for the interpolation method, if the
    method is `normal` it resembles the width of the radial basis
    function used to weigh the points. It is ignored if if the method is
    "increasing" or "decreasing".
    """

    def __init__(self, n_chunks=10, span=1, method="normal"):
        self.span = span
        self.method = method
        self.n_chunks = n_chunks

    def fit(self, X, y):
        """Fits the estimator"""
        allowed_methods = ["average", "normal", "increasing", "decreasing"]
        if self.method not in allowed_methods:
            raise ValueError(
                f"`method` must be in {allowed_methods}, got `{self.method}`"
            )
        if self.n_chunks <= 0:
            raise ValueError(f"`n_chunks` must be >= 1, received {self.n_chunks}")
        if self.span > 1.0:
            raise ValueError(
                f"Error, we expect 0 <= span <= 1, received span={self.span}"
            )
        if self.span < 0.0:
            raise ValueError(
                f"Error, we expect 0 <= span <= 1, received span={self.span}"
            )

        # these two matrices will have shape (columns, quantiles)
        # quantiles indicate where the interval split occurs
        X, y = check_X_y(X, y, estimator=self)
        self.quantiles_ = np.zeros((X.shape[1], self.n_chunks))
        # heights indicate what heights these intervals will have
        self.heights_ = np.zeros((X.shape[1], self.n_chunks))
        self.num_cols_ = X.shape[1]

        average_func = (
            _mk_average
            if self.method in ["average", "normal"]
            else _mk_monotonic_average
        )

        for col in range(X.shape[1]):
            self.quantiles_[col, :] = np.quantile(
                X[:, col], q=np.linspace(0, 1, self.n_chunks)
            )
            self.heights_[col, :] = average_func(
                X[:, col],
                y,
                self.quantiles_[col, :],
                span=self.span,
                method=self.method,
            )
        return self

    def transform(self, X):
        """
        Transform each column such that it is bends smoothly towards y.
        """
        check_is_fitted(self, ["quantiles_", "heights_", "num_cols_"])
        X = check_array(X, estimator=self)
        if X.shape[1] != self.num_cols_:
            raise ValueError(
                f"fitted on {self.num_cols_} features but received {X.shape[1]}"
            )
        transformed = np.zeros(X.shape)
        for col in range(transformed.shape[1]):
            transformed[:, col] = np.interp(
                X[:, col], self.quantiles_[col, :], self.heights_[col, :]
            )
        return transformed



import pytest
import numpy as np
from sklearn.ensemble import IsolationForest
from sklearn.svm import OneClassSVM
from sklearn.linear_model import LinearRegression
from sklearn.neighbors import LocalOutlierFactor

from sklego.common import flatten
from sklego.mixture import GMMOutlierDetector
from sklego.meta import OutlierClassifier

from tests.conftest import general_checks, select_tests


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks]),
        exclude=[
            "check_sample_weights_invariance",
        ]
    )
)
def test_estimator_checks(test_fn):
    mod_quantile = GMMOutlierDetector(threshold=0.999, method="quantile")
    clf_quantile = OutlierClassifier(mod_quantile)
    test_fn('OutlierClassifier', clf_quantile)


@pytest.fixture
def dataset():
    np.random.seed(42)
    return np.random.normal(0, 1, (2000, 2))


@pytest.mark.parametrize('outlier_model', [GMMOutlierDetector(), OneClassSVM(nu=0.05), IsolationForest()])
def test_obvious_usecase(dataset, outlier_model):
    outlier_clf = OutlierClassifier(outlier_model)
    X = dataset
    y = (dataset.max(axis=1) > 3).astype(int)
    outlier_clf.fit(X, y)
    assert outlier_clf.predict([[10, 10]]) == np.array([1])
    assert outlier_clf.predict([[0, 0]]) == np.array([0])
    np.testing.assert_array_almost_equal(outlier_clf.predict_proba([[0, 0]]), np.array([[1, 0]]), decimal=3)
    np.testing.assert_allclose(outlier_clf.predict_proba([[10, 10]]), np.array([[0, 1]]), atol=0.2)
    assert isinstance(outlier_clf.score(X, y), float)


def test_raises_error(dataset):
    mod_quantile = LinearRegression()
    clf_quantile = OutlierClassifier(mod_quantile)
    X = dataset
    y = (dataset.max(axis=1) > 3).astype(int)
    with pytest.raises(ValueError):
        clf_quantile.fit(X, y)


def test_raises_error_no_decision_function(dataset):
    outlier_model = LocalOutlierFactor()
    clf_model = OutlierClassifier(outlier_model)
    X = dataset
    y = (dataset.max(axis=1) > 3).astype(int)
    with pytest.raises(ValueError):
        clf_model.fit(X, y)



KNOWN_PACKAGES = {"cvxpy": {"version": ">=1.0.24", "extra_name": "cvxpy"}}


class NotInstalledPackage:
    """
    Class to gracefully catch ImportErrors for modules and packages that are not installed

    :param package_name (str): Name of the package you want to load
    :param version (str, Optional): Version of the package

    Usage:
        >>> try:
        ...     import thispackagedoesnotexist as package
        >>> except ImportError:
        ...     from sklego.notinstalled import NotInstalledPackage
        ...     package = NotInstalledPackage("thispackagedoesnotexist")
    """

    def __init__(self, package_name: str, version: str = None):
        self.package_name = package_name
        package_info = KNOWN_PACKAGES.get(package_name, {})
        self.version = version if version else package_info.get("version", "")

        extra_name = package_info.get("extra_name", None)
        self.pip_message = (
            (
                f"Install extra requirement {package_name} using "
                + f"`python -m pip install scikit-lego[{extra_name}]` or "
                + "`python -m pip install scikit-lego[all]`. "
                + "For more information, check the 'Dependency installs' section of the installation docs at "
                + "https://scikit-lego.readthedocs.io/en/latest/install.html"
            )
            if extra_name
            else ""
        )

    def __getattr__(self, name):
        raise ImportError(
            f"The package {self.package_name}{self.version} is not installed. "
            + self.pip_message
        )



import numpy as np
from patsy import dmatrix, build_design_matrices, PatsyError
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted


class PatsyTransformer(TransformerMixin, BaseEstimator):
    """
    The patsy transformer offers a method to select the right columns
    from a dataframe as well as a DSL for transformations. It is inspired
    from R formulas.
    This is can be useful as a first step in the pipeline.
    :param formula: a patsy-compatible formula
    :return_type: Either "matrix" or "dataframe", passed on to patsy
    """

    def __init__(self, formula, return_type="matrix"):
        self.formula = formula
        self.return_type = return_type

    def fit(self, X, y=None):
        """Fits the estimator"""
        X_ = dmatrix(self.formula, X, NA_action="raise", return_type=self.return_type)

        # check the number of observations hasn't changed. This ought not to
        # be necessary given NA_action='raise' above but just to be safe
        assert np.array(X_).shape[0] == np.array(X).shape[0]
        self.design_info_ = X_.design_info
        return self

    def transform(self, X):
        """
        Applies the formula to the matrix/dataframe X.

        Returns
        - A patsy.DesignMatrix, if return_type="matrix" (the default)
        - A pandas.DataFrame, if return_type="dataframe"
        """
        check_is_fitted(self, "design_info_")
        try:
            return build_design_matrices(
                [self.design_info_], X, return_type=self.return_type
            )[0]
        except PatsyError as e:
            raise RuntimeError from e



import pytest
import numpy as np
from sklearn.utils import estimator_checks

from sklego.common import flatten
from sklego.preprocessing import IntervalEncoder
from tests.conftest import transformer_checks, general_checks


@pytest.mark.parametrize(
    "test_fn",
    flatten(
        [
            transformer_checks,
            general_checks,
            estimator_checks.check_estimators_dtypes,
            estimator_checks.check_fit_score_takes_y,
            # estimator_checks.check_dtype_object,
            # estimator_checks.check_sample_weights_pandas_series,
            # estimator_checks.check_sample_weights_list,
            # estimator_checks.check_sample_weights_invariance,
            estimator_checks.check_estimators_fit_returns_self,
            estimator_checks.check_complex_data,
            estimator_checks.check_estimators_empty_data_messages,
            estimator_checks.check_pipeline_consistency,
            estimator_checks.check_estimators_nan_inf,
            estimator_checks.check_estimators_overwrite_params,
            estimator_checks.check_estimator_sparse_data,
            estimator_checks.check_estimators_pickle,
        ]
    ),
)
def test_estimator_checks(test_fn):
    test_fn(IntervalEncoder.__name__, IntervalEncoder(n_chunks=2))


@pytest.mark.parametrize("chunks", [1, 2, 5, 10])
def test_obvious_cases_one(random_xy_dataset_regr, chunks):
    X, y = random_xy_dataset_regr
    y = np.ones(y.shape)
    x_transform = IntervalEncoder(n_chunks=chunks).fit(X, y).transform(X)
    assert x_transform.shape == X.shape
    assert np.all(np.isclose(x_transform, 1.0))


@pytest.mark.parametrize("method", ["average", "normal", "increasing", "decreasing"])
def test_obvious_cases_two(random_xy_dataset_regr_small, method):
    X, y = random_xy_dataset_regr_small
    y = np.ones(y.shape)
    x_transform = IntervalEncoder(method=method).fit(X, y).transform(X)
    assert x_transform.shape == X.shape
    assert np.all(np.isclose(x_transform, 1.0))


def generate_dataset(start, n=600):
    np.random.seed(42)
    xs = np.arange(start, start + n) / 100 / np.pi
    y = np.sin(xs) + np.random.normal(0, 0.1, n)
    return xs.reshape(-1, 1), y


@pytest.mark.parametrize("data_init", [50, 600, 1200, 2100])
def test_monotonicity_increasing(data_init):
    X, y = generate_dataset(start=data_init)
    encoder = IntervalEncoder(n_chunks=40, method="increasing")
    y_transformed = encoder.fit_transform(X, y).reshape(-1).round(4)
    for i in range(len(y_transformed) - 1):
        assert y_transformed[i] <= y_transformed[i + 1]


@pytest.mark.parametrize("data_init", [50, 600, 1200, 2100])
def test_monotonicity_decreasing(data_init):
    X, y = generate_dataset(start=data_init)
    encoder = IntervalEncoder(n_chunks=40, method="decreasing")
    y_transformed = encoder.fit_transform(X, y).reshape(-1).round(4)
    print(y_transformed.reshape(-1))
    for i in range(len(y_transformed) - 1):
        assert y_transformed[i] >= y_transformed[i + 1]


def test_throw_valuerror_given_nonsense():
    X = np.ones((10, 2))
    y = np.ones(10)
    with pytest.raises(ValueError):
        IntervalEncoder(n_chunks=0).fit(X, y)
    with pytest.raises(ValueError):
        IntervalEncoder(n_chunks=-1).fit(X, y)
    with pytest.raises(ValueError):
        IntervalEncoder(span=-0.1).fit(X, y)
    with pytest.raises(ValueError):
        IntervalEncoder(span=2.0).fit(X, y)
    with pytest.raises(ValueError):
        IntervalEncoder(method="dinosaurhead").fit(X, y)



import numpy as np
import pytest
from sklearn.cluster import DBSCAN
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import Ridge, LogisticRegression

from sklego.common import flatten
from sklego.meta import SubjectiveClassifier
from tests.conftest import general_checks, classifier_checks, select_tests


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, classifier_checks]),
        exclude=['check_sample_weights_invariance']
        # outliers train wont work because we have two thresholds
    )
)
def test_estimator_checks_classification(test_fn):
    if test_fn.__name__ == "check_classifiers_classes":
        prior = {
            "one": 0.1,
            "two": 0.1,
            "three": 0.1,
            -1: 0.1,
            1: 0.6,
        }  # nonsensical prior to make sklearn check pass
    else:
        prior = {0: 0.7, 1: 0.2, 2: 0.1}

    # Some of the sklearn checkers generate random y data with 3 classes, so prior needs to have these classes
    estimator = SubjectiveClassifier(LogisticRegression(), prior)
    test_fn(SubjectiveClassifier.__name__, estimator)


@pytest.mark.parametrize(
    "classes, prior, cfm, first_class_posterior",
    [
        ([0, 1], [0.8, 0.2], [[90, 10], [10, 90]], 0.973),  # numeric classes
        (["a", "b"], [0.8, 0.2], [[90, 10], [10, 90]], 0.973),  # char classes
        ([False, True], [0.8, 0.2], [[90, 10], [10, 90]], 0.973),  # bool classes
        (
            ["a", "b", "c"],
            [0.8, 0.1, 0.1],
            [[80, 10, 10], [10, 90, 0], [0, 0, 100]],
            0.985,
        ),  # n classes
        (
            [0, 1],
            [0.8, 0.2],
            [[100, 0], [0, 100]],
            1.0,
        ),  # "perfect" confusion matrix (no FP) -> prior is ignored
        (
            [0, 1],
            [0.2, 0.8],
            [[0, 100], [0, 100]],
            0.2,
        ),  # failure to predict class by inner estimator
        (
            [0, 1, 2],
            [0.1, 0.1, 0.8],
            [[0, 0, 100], [0, 0, 100], [0, 0, 100]],
            0.1,
        ),  # extremely biased, n classes
        (
            [0, 1, 2],
            [0.2, 0.1, 0.7],
            [[80, 0, 20], [0, 0, 100], [10, 0, 90]],
            0.696,
        ),  # biased, n classes
    ],
)
def test_posterior_computation(mocker, classes, prior, cfm, first_class_posterior):
    def mock_confusion_matrix(y, y_pred):
        return np.array(cfm)

    mocker.patch("sklego.meta.subjective_classifier.confusion_matrix", side_effect=mock_confusion_matrix)
    mock_estimator = mocker.Mock(RandomForestClassifier())
    mock_estimator.classes_ = np.array(classes)
    subjective_model = SubjectiveClassifier(mock_estimator, dict(zip(classes, prior)))
    subjective_model.fit(np.zeros((10, 10)), np.array([classes[0]] * 10))
    assert (
        pytest.approx(subjective_model.posterior_matrix_[0, 0], 0.001)
        == first_class_posterior
    )
    assert np.isclose(subjective_model.posterior_matrix_.sum(axis=0), 1).all()


@pytest.mark.parametrize(
    "prior, y",
    [
        (
            {"a": 0.8, "b": 0.2},
            ["a", "c"],
        ),  # class from train data not defined in prior
        ({"a": 0.8, "b": 0.2}, [0, 1]),  # different data types
    ],
)
def test_fit_y_data_inconsistent_with_prior_failure_conditions(prior, y):
    with pytest.raises(ValueError) as exc:
        SubjectiveClassifier(RandomForestClassifier(), prior).fit(
            np.zeros((len(y), 2)), np.array(y)
        )

    assert str(exc.value).startswith("Training data is inconsistent with prior")


def test_to_discrete():
    assert np.isclose(
        SubjectiveClassifier._to_discrete(
            np.array([[1, 0], [0.8, 0.2], [0.5, 0.5], [0.2, 0.8]])
        ),
        np.array([[1, 0], [1, 0], [1, 0], [0, 1]]),
    ).all()


@pytest.mark.parametrize(
    "weights,y_hats,expected_probas",
    [
        (
            [0.8, 0.2],
            [[1, 0], [0.5, 0.5], [0.8, 0.2]],
            [[1, 0], [0.8, 0.2], [0.94, 0.06]],
        ),
        (
            [0.5, 0.5],
            [[1, 0], [0.5, 0.5], [0.8, 0.2]],
            [[1, 0], [0.5, 0.5], [0.8, 0.2]],
        ),
        ([[0.8, 0.2], [0.5, 0.5]], [[1, 0], [0.8, 0.2]], [[1, 0], [0.8, 0.2]]),
    ],
)
def test_weighted_proba(weights, y_hats, expected_probas):
    assert np.isclose(
        SubjectiveClassifier._weighted_proba(np.array(weights), np.array(y_hats)),
        np.array(expected_probas),
        atol=1e-02,
    ).all()


@pytest.mark.parametrize(
    "evidence_type,expected_probas",
    [
        ("predict_proba", [[0.94, 0.06], [1, 0], [0.8, 0.2], [0.5, 0.5]]),
        ("confusion_matrix", [[0.97, 0.03], [0.97, 0.03], [0.97, 0.03], [0.47, 0.53]]),
        ("both", [[0.99, 0.01], [1, 0], [0.97, 0.03], [0.18, 0.82]]),
    ],
)
def test_predict_proba(mocker, evidence_type, expected_probas):
    def mock_confusion_matrix(y, y_pred):
        return np.array([[80, 20], [10, 90]])

    mocker.patch("sklego.meta.subjective_classifier.confusion_matrix", side_effect=mock_confusion_matrix)
    mock_inner_estimator = mocker.Mock(RandomForestClassifier)
    mock_inner_estimator.predict_proba.return_value = np.array(
        [[0.8, 0.2], [1, 0], [0.5, 0.5], [0.2, 0.8]]
    )
    mock_inner_estimator.classes_ = np.array([0, 1])
    subjective_model = SubjectiveClassifier(
        mock_inner_estimator, {0: 0.8, 1: 0.2}, evidence=evidence_type
    )
    subjective_model.fit(np.zeros((10, 10)), np.zeros(10))
    posterior_probabilities = subjective_model.predict_proba(np.zeros((4, 2)))
    assert posterior_probabilities.shape == (4, 2)
    assert np.isclose(
        posterior_probabilities, np.array(expected_probas), atol=0.01
    ).all()


@pytest.mark.parametrize(
    "inner_estimator, prior, evidence, expected_error_msg",
    [
        (DBSCAN(), {"a": 1}, "both", "Invalid inner estimator"),
        (Ridge(), {"a": 1}, "predict_proba", "Invalid inner estimator"),
        (
            RandomForestClassifier(),
            {"a": 0.8, "b": 0.1},
            "confusion_matrix",
            "Invalid prior",
        ),
        (
            RandomForestClassifier(),
            {"a": 0.8, "b": 0.2},
            "foo_evidence",
            "Invalid evidence",
        ),
    ],
)
def test_params_failure_conditions(
    inner_estimator, prior, evidence, expected_error_msg
):
    with pytest.raises(ValueError) as exc:
        SubjectiveClassifier(inner_estimator, prior, evidence).fit(
            np.zeros((2, 2)), np.zeros(2)
        )

    assert str(exc.value).startswith(expected_error_msg)



import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.mixture import GaussianMixture, BayesianGaussianMixture
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted, check_array, FLOAT_DTYPES


class GaussianMixtureNB(BaseEstimator, ClassifierMixin):
    """
    The GaussianMixtureNB trains a Naive Bayes Classifier that uses a mixture
    of gaussians instead of merely training a single one.

    You can pass any keyword parameter that scikit-learn's Gaussian Mixture
    Model uses and it will be passed along.
    """

    def __init__(
        self,
        n_components=1,
        covariance_type="full",
        tol=1e-3,
        reg_covar=1e-6,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weights_init=None,
        means_init=None,
        precisions_init=None,
        random_state=None,
        warm_start=False,
    ):
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.weights_init = weights_init
        self.means_init = means_init
        self.precisions_init = precisions_init
        self.random_state = random_state
        self.warm_start = warm_start

    def fit(self, X: np.array, y: np.array) -> "GaussianMixtureNB":
        """
        Fit the model using X, y as training data.

        :param X: array-like, shape=(n_columns, n_samples, ) training data.
        :param y: array-like, shape=(n_samples, ) training data.
        :return: Returns an instance of self.
        """
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)
        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        self.gmms_ = {}
        self.classes_ = unique_labels(y)
        self.num_fit_cols_ = X.shape[1]
        for c in self.classes_:
            subset_x, subset_y = X[y == c], y[y == c]
            self.gmms_[c] = [
                GaussianMixture(
                    n_components=self.n_components,
                    covariance_type=self.covariance_type,
                    tol=self.tol,
                    reg_covar=self.reg_covar,
                    max_iter=self.max_iter,
                    n_init=self.n_init,
                    init_params=self.init_params,
                    weights_init=self.weights_init,
                    means_init=self.means_init,
                    precisions_init=self.precisions_init,
                    random_state=self.random_state,
                    warm_start=self.warm_start,
                ).fit(subset_x[:, i].reshape(-1, 1), subset_y)
                for i in range(X.shape[1])
            ]
        return self

    def predict(self, X):
        check_is_fitted(self, ["gmms_", "classes_"])
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    def predict_proba(self, X: np.array):
        check_is_fitted(self, ["gmms_", "classes_"])
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        if self.num_fit_cols_ != X.shape[1]:
            raise ValueError(
                f"number of columns {X.shape[1]} does not match fit size {self.num_fit_cols_}"
            )
        check_is_fitted(self, ["gmms_", "classes_"])
        probs = np.zeros((X.shape[0], len(self.classes_)))
        for k, v in self.gmms_.items():
            class_idx = int(np.argwhere(self.classes_ == k))
            probs[:, class_idx] = np.array(
                [
                    m.score_samples(np.expand_dims(X[:, idx], 1))
                    for idx, m in enumerate(v)
                ]
            ).sum(axis=0)
        likelihood = np.exp(probs)
        return likelihood / likelihood.sum(axis=1).reshape(-1, 1)


class BayesianGaussianMixtureNB(BaseEstimator, ClassifierMixin):
    """
    The BayesianGaussianMixtureNB trains a Naive Bayes Classifier that uses a bayesian
    mixture of gaussians instead of merely training a single one.

    You can pass any keyword parameter that scikit-learn's Bayesian Gaussian Mixture
    Model uses and it will be passed along.
    """

    def __init__(
        self,
        n_components=1,
        covariance_type="full",
        tol=0.001,
        reg_covar=1e-06,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weight_concentration_prior_type="dirichlet_process",
        weight_concentration_prior=None,
        mean_precision_prior=None,
        mean_prior=None,
        degrees_of_freedom_prior=None,
        covariance_prior=None,
        random_state=None,
        warm_start=False,
        verbose=0,
        verbose_interval=10,
    ):
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior
        self.mean_precision_prior = mean_precision_prior
        self.mean_prior = mean_prior
        self.degrees_of_freedom_prior = degrees_of_freedom_prior
        self.covariance_prior = covariance_prior
        self.random_state = random_state
        self.warm_start = warm_start
        self.verbose = verbose
        self.verbose_interval = verbose_interval

    def fit(self, X: np.array, y: np.array) -> "BayesianGaussianMixtureNB":
        """
        Fit the model using X, y as training data.

        :param X: array-like, shape=(n_columns, n_samples, ) training data.
        :param y: array-like, shape=(n_samples, ) training data.
        :return: Returns an instance of self.
        """
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)
        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        self.gmms_ = {}
        self.classes_ = unique_labels(y)
        self.num_fit_cols_ = X.shape[1]
        for c in self.classes_:
            subset_x, subset_y = X[y == c], y[y == c]
            self.gmms_[c] = [
                BayesianGaussianMixture(
                    n_components=self.n_components,
                    covariance_type=self.covariance_type,
                    tol=self.tol,
                    reg_covar=self.reg_covar,
                    max_iter=self.max_iter,
                    n_init=self.n_init,
                    init_params=self.init_params,
                    weight_concentration_prior_type=self.weight_concentration_prior_type,
                    weight_concentration_prior=self.weight_concentration_prior,
                    mean_precision_prior=self.mean_precision_prior,
                    mean_prior=self.mean_prior,
                    degrees_of_freedom_prior=self.degrees_of_freedom_prior,
                    covariance_prior=self.covariance_prior,
                    random_state=self.random_state,
                    warm_start=self.warm_start,
                    verbose=self.verbose,
                    verbose_interval=self.verbose_interval,
                ).fit(subset_x[:, i].reshape(-1, 1), subset_y)
                for i in range(X.shape[1])
            ]
        return self

    def predict(self, X):
        check_is_fitted(self, ["gmms_", "classes_", "num_fit_cols_"])
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    def predict_proba(self, X: np.array):
        check_is_fitted(self, ["gmms_", "classes_", "num_fit_cols_"])
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        if self.num_fit_cols_ != X.shape[1]:
            raise ValueError(
                f"number of columns {X.shape[1]} does not match fit size {self.num_fit_cols_}"
            )
        check_is_fitted(self, ["gmms_", "classes_"])
        probs = np.zeros((X.shape[0], len(self.classes_)))
        for k, v in self.gmms_.items():
            class_idx = int(np.argwhere(self.classes_ == k))
            probs[:, class_idx] = np.array(
                [
                    m.score_samples(np.expand_dims(X[:, idx], 1))
                    for idx, m in enumerate(v)
                ]
            ).sum(axis=0)
        likelihood = np.exp(probs)
        return likelihood / likelihood.sum(axis=1).reshape(-1, 1)



import numpy as np
import pytest

from sklego.common import flatten
from sklego.preprocessing import IdentityTransformer

from tests.conftest import select_tests, transformer_checks, general_checks, nonmeta_checks


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, transformer_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
def test_estimator_checks(test_fn):
    test_fn(IdentityTransformer.__name__, IdentityTransformer(check_X=True))


def test_same_values(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    X_new = IdentityTransformer(check_X=True).fit_transform(X)
    assert np.isclose(X, X_new).all()


def test_nan_inf(random_xy_dataset_regr):
    # see https://github.com/koaning/scikit-lego/pull/527
    X, y = random_xy_dataset_regr
    X = X.astype(np.float32)
    X[np.random.ranf(size=X.shape) > 0.9] = np.nan
    X[np.random.ranf(size=X.shape) > 0.9] = -np.inf
    X[np.random.ranf(size=X.shape) > 0.9] = np.inf
    X_new = IdentityTransformer(check_X=False).fit_transform(X)



import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, TransformerMixin, clone
from sklearn.utils.validation import check_is_fitted

from ._grouped_utils import _split_groups_and_values


class GroupedTransformer(BaseEstimator, TransformerMixin):
    """
    Construct a transformer per data group. Splits data by groups from single or multiple columns
    and transforms remaining columns using the transformers corresponding to the groups.

    :param transformer: the transformer to be applied per group
    :param groups: the column(s) of the matrix/dataframe to select as a grouping parameter set. If None,
                   the transformer will be applied to the entire input without grouping
    :param use_global_model: Whether or not to fall back to a general transformation in case a group
                             is not found during `.transform()`
    """

    _check_kwargs = {"accept_large_sparse": False}

    def __init__(self, transformer, groups, use_global_model=True):
        self.transformer = transformer
        self.groups = groups
        self.use_global_model = use_global_model

    def __fit_single_group(self, group, X, y=None):
        try:
            return clone(self.transformer).fit(X, y)
        except Exception as e:
            raise type(e)(f"Exception for group {group}: {e}")

    def __fit_grouped_transformer(
        self, X_group: pd.DataFrame, X_value: np.array, y=None
    ):
        """Fit a transformer to each group"""
        # Make the groups based on the groups dataframe, use the indices on the values array
        try:
            group_indices = X_group.groupby(X_group.columns.tolist()).indices
        except TypeError:
            # This one is needed because of line #918 of sklearn/utils/estimator_checks
            raise TypeError("argument must be a string, date or number")

        if y is not None:
            if isinstance(y, pd.Series):
                y.index = X_group.index

            grouped_transformers = {
                # Fit a clone of the transformer to each group
                group: self.__fit_single_group(group, X_value[indices, :], y[indices])
                for group, indices in group_indices.items()
            }
        else:
            grouped_transformers = {
                group: self.__fit_single_group(group, X_value[indices, :])
                for group, indices in group_indices.items()
            }

        return grouped_transformers

    def __check_transformer(self):
        if not hasattr(self.transformer, "transform"):
            raise ValueError(
                "The supplied transformer should have a 'transform' method"
            )

    def fit(self, X, y=None):
        """
        Fit the transformers to the groups in X

        :param X: Array-like with at least two columns, of which at least one corresponds to groups defined in init,
                  and the remaining columns represent the values to transform.
        :param y: (Optional) target variable
        """
        self.__check_transformer()

        self.fallback_ = None

        if self.groups is None:
            self.transformers_ = clone(self.transformer).fit(X, y)
            return self

        X_group, X_value = _split_groups_and_values(
            X, self.groups, **self._check_kwargs
        )
        self.transformers_ = self.__fit_grouped_transformer(X_group, X_value, y)

        if self.use_global_model:
            self.fallback_ = clone(self.transformer).fit(X_value)

        return self

    def __transform_single_group(self, group, X):
        """Transform a single group by getting its transformer from the fitted dict"""
        # Keep track of the original index such that we can sort in __transform_groups
        index = X.index
        try:
            group_transformer = self.transformers_[group]
        except KeyError:
            if self.fallback_:
                group_transformer = self.fallback_
            else:
                raise ValueError(
                    f"Found new group {group} during transform with use_global_model = False"
                )

        return pd.DataFrame(group_transformer.transform(X)).set_index(index)

    def __transform_groups(self, X_group: pd.DataFrame, X_value: np.array):
        """Transform all groups"""
        # Reset indices such that they are the same in X_group (reset in __check_grouping_columns),
        # this way we can track the order of the result
        X_value = pd.DataFrame(X_value).reset_index(drop=True)

        # Make the groups based on the groups dataframe, use the indices on the values array
        group_indices = X_group.groupby(X_group.columns.tolist()).indices

        return (
            pd.concat(
                [
                    self.__transform_single_group(group, X_value.loc[indices, :])
                    for group, indices in group_indices.items()
                ],
                axis=0,
            )
            .sort_index()
            .values
        )

    def transform(self, X):
        """
        Fit the transformers to the groups in X

        :param X: Array-like with columns corresponding to the ones in .fit()
        """
        check_is_fitted(self, ["fallback_", "transformers_"])

        if self.groups is None:
            return self.transformers_.transform(X)

        X_group, X_value = _split_groups_and_values(
            X, self.groups, **self._check_kwargs
        )

        return self.__transform_groups(X_group, X_value)



"""Test the ZeroInflatedRegressor."""

import numpy as np
import pytest
from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

from sklego.common import flatten
from sklego.meta import ZeroInflatedRegressor
from sklego.testing import check_shape_remains_same_regressor
from tests.conftest import general_checks, select_tests, regressor_checks


@pytest.mark.parametrize("test_fn", [check_shape_remains_same_regressor])
def test_zir(test_fn):
    regr = ZeroInflatedRegressor(
        classifier=ExtraTreesClassifier(random_state=0),
        regressor=ExtraTreesRegressor(random_state=0)
    )
    test_fn(ZeroInflatedRegressor.__name__, regr)


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, regressor_checks]),
    )
)
def test_estimator_checks(test_fn):
    test_fn(
        ZeroInflatedRegressor.__name__,
        ZeroInflatedRegressor(
            classifier=ExtraTreesClassifier(random_state=0),
            regressor=ExtraTreesRegressor(random_state=0)
        )
    )


def test_zero_inflated_example():
    from sklearn.model_selection import cross_val_score

    np.random.seed(0)
    X = np.random.randn(10000, 4)
    y = ((X[:, 0] > 0) & (X[:, 1] > 0)) * np.abs(X[:, 2] * X[:, 3] ** 2)  # many zeroes here, in about 75% of the cases.

    zir = ZeroInflatedRegressor(
        classifier=ExtraTreesClassifier(random_state=0),
        regressor=ExtraTreesRegressor(random_state=0)
    )

    zir_score = cross_val_score(zir, X, y).mean()
    et_score = cross_val_score(ExtraTreesRegressor(), X, y).mean()

    assert zir_score > 0.85
    assert zir_score > et_score


@pytest.mark.parametrize("classifier,regressor,performance", [
    (ExtraTreesClassifier(random_state=0), ExtraTreesRegressor(random_state=0), 0.85),
    (KNeighborsClassifier(), KNeighborsRegressor(), 0.55)
])
def test_zero_inflated_with_sample_weights_example(classifier, regressor, performance):
    from sklearn.model_selection import cross_val_score

    np.random.seed(0)
    X = np.random.randn(10000, 4)
    y = ((X[:, 0] > 0) & (X[:, 1] > 0)) * np.abs(X[:, 2] * X[:, 3] ** 2)  # many zeroes here, in about 75% of the cases.

    zir = ZeroInflatedRegressor(
        classifier=classifier,
        regressor=regressor
    )

    zir_score = cross_val_score(zir, X, y, fit_params={'sample_weight': np.arange(len(y))}).mean()

    assert zir_score > performance


def test_wrong_estimators_exceptions():
    X = np.array([[0.]])
    y = np.array([0.])

    with pytest.raises(ValueError, match="`classifier` has to be a classifier."):
        zir = ZeroInflatedRegressor(ExtraTreesRegressor(), ExtraTreesRegressor())
        zir.fit(X, y)

    with pytest.raises(ValueError, match="`regressor` has to be a regressor."):
        zir = ZeroInflatedRegressor(ExtraTreesClassifier(), ExtraTreesClassifier())
        zir.fit(X, y)



import pytest
import numpy as np

from sklego.common import flatten
from sklego.naive_bayes import GaussianMixtureNB, BayesianGaussianMixtureNB
from tests.conftest import general_checks, nonmeta_checks, select_tests


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_non_transformer_estimators_n_iter",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
def test_estimator_checks(test_fn):
    clf1 = GaussianMixtureNB()
    clf2 = GaussianMixtureNB(n_components=2)
    clf3 = BayesianGaussianMixtureNB()
    clf4 = BayesianGaussianMixtureNB(n_components=2)
    test_fn(GaussianMixtureNB.__name__, clf1)
    test_fn(GaussianMixtureNB.__name__ + "_components_5", clf2)
    test_fn(BayesianGaussianMixtureNB.__name__, clf3)
    test_fn(BayesianGaussianMixtureNB.__name__ + "_components_5", clf4)


@pytest.fixture
def dataset():
    np.random.seed(42)
    return np.concatenate([np.random.normal(0, 1, (2000, 2))])


@pytest.mark.parametrize("k", [1, 5, 10])
def test_obvious_usecase(k):
    X = np.concatenate(
        [np.random.normal(-10, 1, (100, 2)), np.random.normal(10, 1, (100, 2))]
    )
    y = np.concatenate([np.zeros(100), np.ones(100)])
    assert (GaussianMixtureNB(n_components=k, max_iter=1000).fit(X, y).predict(X) == y).all()
    assert (BayesianGaussianMixtureNB(n_components=k, max_iter=1000).fit(X, y).predict(X) == y).all()



import numpy as np
import warnings

from typing import Callable


def correlation_score(column):
    """
    The correlation score can score how well the estimator predictions correlate with a given column.
    This is especially useful to use in situations where "fairness" is a theme.

    `correlation_score` takes a column on which to calculate the correlation and returns a metric function

    Usage:
    `correlation_score('gender')(clf, X, y)`


    :param column: Name of the column (when X is a dataframe) or the index of the column (when X is a numpy array).
    :return:
        A function which calculates the negative correlation between estimator.predict(X) and X[column]
        (in gridsearch, larger is better and we want to typically punish correlation).
    """

    def correlation_metric(estimator, X, y_true=None):
        """Remember: X is the thing going *in* to your pipeline."""
        sensitive_col = X[:, column] if isinstance(X, np.ndarray) else X[column]
        return -np.abs(np.corrcoef(estimator.predict(X), sensitive_col)[1, 0])

    return correlation_metric


def p_percent_score(sensitive_column, positive_target=1):
    r"""
    The p_percent score calculates the ratio between the probability of a positive outcome
    given the sensitive attribute (column) being true and the same probability given the
    sensitive attribute being false.

    .. math::
        \min \left(\frac{P(\hat{y}=1 | z=1)}{P(\hat{y}=1 | z=0)}, \frac{P(\hat{y}=1 | z=0)}{P(\hat{y}=1 | z=1)}\right)

    This is especially useful to use in situations where "fairness" is a theme.

    Usage:
    `p_percent_score('gender')(clf, X, y)`

    source:
    - M. Zafar et al. (2017), Fairness Constraints: Mechanisms for Fair Classification

    :param sensitive_column:
        Name of the column containing the binary sensitive attribute (when X is a dataframe)
        or the index of the column (when X is a numpy array).
    :param positive_target: The name of the class which is associated with a positive outcome
    :return: a function (clf, X, y_true) -> float that calculates the p percent score for z = column
    """

    def impl(estimator, X, y_true=None):
        """Remember: X is the thing going *in* to your pipeline."""
        sensitive_col = (
            X[:, sensitive_column] if isinstance(X, np.ndarray) else X[sensitive_column]
        )

        if not np.all((sensitive_col == 0) | (sensitive_col == 1)):
            raise ValueError(
                f"p_percent_score only supports binary indicator columns for `column`. "
                f"Found values {np.unique(sensitive_col)}"
            )

        y_hat = estimator.predict(X)
        y_given_z1 = y_hat[sensitive_col == 1]
        y_given_z0 = y_hat[sensitive_col == 0]
        p_y1_z1 = np.mean(y_given_z1 == positive_target)
        p_y1_z0 = np.mean(y_given_z0 == positive_target)

        # If we never predict a positive target for one of the subgroups, the model is by definition not
        # fair so we return 0
        if p_y1_z1 == 0:
            warnings.warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 1, returning 0",
                RuntimeWarning,
            )
            return 0

        if p_y1_z0 == 0:
            warnings.warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 0, returning 0",
                RuntimeWarning,
            )
            return 0

        p_percent = np.minimum(p_y1_z1 / p_y1_z0, p_y1_z0 / p_y1_z1)
        return p_percent if not np.isnan(p_percent) else 1

    return impl


def equal_opportunity_score(sensitive_column, positive_target=1):
    r"""
    The equality opportunity score calculates the ratio between the probability of a **true positive** outcome
    given the sensitive attribute (column) being true and the same probability given the
    sensitive attribute being false.

    .. math::
        \min \left(\frac{P(\hat{y}=1 | z=1, y=1)}{P(\hat{y}=1 | z=0, y=1)},
        \frac{P(\hat{y}=1 | z=0, y=1)}{P(\hat{y}=1 | z=1, y=1)}\right)

    This is especially useful to use in situations where "fairness" is a theme.

    Usage:
    `equal_opportunity_score('gender')(clf, X, y)`

    Source:
    - M. Hardt, E. Price and N. Srebro (2016), Equality of Opportunity in Supervised Learning

    :param sensitive_column:
        Name of the column containing the binary sensitive attribute (when X is a dataframe)
        or the index of the column (when X is a numpy array).
    :param positive_target: The name of the class which is associated with a positive outcome
    :return: a function (clf, X, y_true) -> float that calculates the equal opportunity score for z = column
    """

    def impl(estimator, X, y_true):
        """Remember: X is the thing going *in* to your pipeline."""
        sensitive_col = (
            X[:, sensitive_column] if isinstance(X, np.ndarray) else X[sensitive_column]
        )

        if not np.all((sensitive_col == 0) | (sensitive_col == 1)):
            raise ValueError(
                f"equal_opportunity_score only supports binary indicator columns for `column`. "
                f"Found values {np.unique(sensitive_col)}"
            )

        y_hat = estimator.predict(X)
        y_given_z1_y1 = y_hat[(sensitive_col == 1) & (y_true == positive_target)]
        y_given_z0_y1 = y_hat[(sensitive_col == 0) & (y_true == positive_target)]

        # If we never predict a positive target for one of the subgroups, the model is by definition not
        # fair so we return 0
        if len(y_given_z1_y1) == 0:
            warnings.warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 1, returning 0",
                RuntimeWarning,
            )
            return 0

        if len(y_given_z0_y1) == 0:
            warnings.warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 0, returning 0",
                RuntimeWarning,
            )
            return 0

        p_y1_z1 = np.mean(y_given_z1_y1 == positive_target)
        p_y1_z0 = np.mean(y_given_z0_y1 == positive_target)
        score = np.minimum(p_y1_z1 / p_y1_z0, p_y1_z0 / p_y1_z1)
        return score if not np.isnan(score) else 1

    return impl


def subset_score(subset_picker: Callable, score: Callable, **kwargs):
    r"""
    Returns a method that applies the passed score only to a specific subset. The subset picker
    is a method that is passed the corresponding X and y_true and returns a one-dimensional
    boolean vector where every element corresponds to a row in the data. Only the elements
    with a True value are taken into account for the passed score, representing a filter.

    This allows users to have an easy approach to measuring metrics over different slices of
    the population which can give insights into the model performance, either specifically for
    fairness or in general.

    Usage:
    `subset_score(lambda X, y_true: X['column'] == 'A', accuracy_score)(clf, X, y)`

    :param subset_picker: Method that returns a boolean mask that is used for slicing the samples
    :param score: The score that needs to be applied to the subset
    :param kwargs: Additional keyword arguments to pass to score
    :return: a function that calculates the passed score for the subset
    """

    def sliced_metric(estimator, X, y_true=None):
        mask = subset_picker(X, y_true)
        if isinstance(mask, np.ndarray):
            if len(mask.shape) > 1:
                raise ValueError(
                    "`subset_picker` should return 1-dimensional numpy array or Pandas"
                    + " series, returned {} instead".format(len(mask.shape))
                )
        if np.sum(mask) == 0:
            warnings.warn("No samples in subset, returning NaN", RuntimeWarning)
            return np.nan
        X = X[mask]
        y_pred = estimator.predict(X)
        return score(y_true[mask], y_pred, **kwargs)

    return sliced_metric



import numbers
from datetime import timedelta
from itertools import combinations
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.model_selection._split import _BaseKFold, check_array
from sklearn.utils.validation import indexable

from sklego.base import Clusterer
from sklego.common import sliding_window


class TimeGapSplit:
    """
    Provides train/test indices to split time series data samples.
    This cross-validation object is a variation of TimeSeriesSplit with the following 
    differences:

    - The splits are made based on datetime duration, instead of number of rows.
    - The user specifies the validation durations and either training_duration or n_splits.
    - The user can specify a 'gap' duration that is added after the training split and before the validation split.

    The 3 duration parameters can be used to really replicate how the model is going to
    be used in production in batch learning.

    Each validation fold doesn't overlap. The entire 'window' moves by 1 'valid_duration'
    until there is not enough data.

    If this would lead to more splits then specified with 'n_splits', the 'window' moves by
    the 'validation_duration' times the fraction of possible splits and requested splits:

    - n_possible_splits = (total_length - train_duration-gap_duration) // valid_duration
    - time_shift = valid_duration * n_possible_splits / n_slits
    so the CV spans the whole dataset.

    If 'train_duration' is not passed but 'n_splits' is, the training duration is increased to:

        train_duration = total_length - (gap_duration + valid_duration * n_splits)

    such that the shifting the entire window by one validation duration spans the whole training set.

    :param pandas.Series date_serie: series with the date, that should have all the indices of X used in the split() method.
    :param datetime.timedelta train_duration: historical training data.
    :param datetime.timedelta valid_duration: retraining period.
    :param datetime.timedelta gap_duration: forward looking window of the target.
        The period of the forward looking window necessary to create your target variable.
        This period is dropped at the end of your training folds due to lack of recent data.
        In production you would have not been able to create the target for that period, and you would have drop it from
        the training data.
    :param int n_splits: number of splits.
    :param string window: either 'rolling' or 'expanding'.
        'rolling' window has fixed size and is shifted entirely.
        'expanding' left side of window is fixed, right border increases each fold.
    """

    def __init__(
        self,
        date_serie,
        valid_duration,
        train_duration=None,
        gap_duration=timedelta(0),
        n_splits=None,
        window="rolling",
    ):
        if (train_duration is None) and (n_splits is None):
            raise ValueError("Either train_duration or n_splits have to be defined")

        if (train_duration is not None) and (train_duration <= gap_duration):
            raise ValueError(
                "gap_duration is longer than train_duration, it should be shorter."
            )

        if not date_serie.index.is_unique:
            raise ValueError("date_serie doesn't have a unique index")

        self.date_serie = date_serie.copy()
        self.date_serie = self.date_serie.rename("__date__")
        self.train_duration = train_duration
        self.valid_duration = valid_duration
        self.gap_duration = gap_duration
        self.n_splits = n_splits
        self.window = window

    def _join_date_and_x(self, X):
        """
        Make a DataFrame indexed by the pandas index (the same as date_series) with date column joined with that index
        and with the 'numpy index' column (i.e. just a range) that is required for the output and the rest of sklearn.

        :param pandas.DataFrame X: Dataframe with the data to split
        """
        X_index_df = pd.DataFrame(range(len(X)), columns=["np_index"], index=X.index)
        X_index_df = X_index_df.join(self.date_serie)

        return X_index_df

    def split(self, X, y=None, groups=None):
        """
        Generate indices to split data into training and test set.

        :param pandas.DataFrame X: Dataframe with the data to split
        :param y: Always ignored, exists for compatibility
        :param groups: Always ignored, exists for compatibility
        """

        X_index_df = self._join_date_and_x(X)
        X_index_df = X_index_df.sort_values("__date__", ascending=True)

        if len(X) != len(X_index_df):
            raise AssertionError(
                "X and X_index_df are not the same length, "
                "there must be some index missing in 'self.date_serie'"
            )

        date_min = X_index_df["__date__"].min()
        date_max = X_index_df["__date__"].max()
        date_length = X_index_df["__date__"].max() - X_index_df["__date__"].min()

        if (self.train_duration is None) and (self.n_splits is not None):
            self.train_duration = date_length - (
                self.gap_duration + self.valid_duration * self.n_splits
            )

        if (self.train_duration is not None) and (
            self.train_duration <= self.gap_duration
        ):
            raise ValueError(
                "gap_duration is longer than train_duration, it should be shorter."
            )

        n_split_max = (
            date_length - self.train_duration - self.gap_duration
        ) / self.valid_duration
        if self.n_splits:
            if n_split_max < self.n_splits:
                raise ValueError(
                    (
                        "Number of folds requested = {1} are greater"
                        " than maximum  ={0} possible without"
                        " overlapping validation sets."
                    ).format(n_split_max, self.n_splits)
                )

        current_date = date_min
        start_date = date_min
        # if the n_splits is smaller than what would usually be done for train val and gap duration,
        # the next fold is slightly further in time than just valid_duration
        if self.n_splits is not None:
            time_shift = self.valid_duration * n_split_max / self.n_splits
        else:
            time_shift = self.valid_duration
        while True:
            if (
                current_date + self.train_duration + time_shift + self.gap_duration
                > date_max
            ):
                break

            X_train_df = X_index_df[
                (X_index_df["__date__"] >= start_date)
                & (X_index_df["__date__"] < current_date + self.train_duration)
            ]
            X_valid_df = X_index_df[
                (
                    X_index_df["__date__"]
                    >= current_date + self.train_duration + self.gap_duration
                )
                & (
                    X_index_df["__date__"]
                    < current_date
                    + self.train_duration
                    + self.valid_duration
                    + self.gap_duration
                )
            ]

            current_date = current_date + time_shift
            if self.window == "rolling":
                start_date = current_date
            yield (
                X_train_df["np_index"].values,
                X_valid_df["np_index"].values,
            )

    def get_n_splits(self, X=None, y=None, groups=None):
        """Gets the number of splits

        :return: amount of n_splits
        :rtype: int
        """
        return sum(1 for x in self.split(X, y, groups))

    def summary(self, X):
        """Describes all folds

        :param pandas.DataFrame X: Dataframe with the data to split
        :returns: ``pd.DataFrame`` summary of all folds
        """
        summary = []
        X_index_df = self._join_date_and_x(X)

        def get_split_info(X, indices, j, part, summary):
            dates = X_index_df.iloc[indices]["__date__"]
            mindate = dates.min()
            maxdate = dates.max()

            s = pd.Series(
                {
                    "Start date": mindate,
                    "End date": maxdate,
                    "Period": pd.to_datetime(maxdate, format="%Y%m%d")
                    - pd.to_datetime(mindate, format="%Y%m%d"),
                    "Unique days": len(dates.unique()),
                    "nbr samples": len(indices),
                },
                name=(j, part),
            )
            summary.append(s)
            return summary

        j = 0
        for i in self.split(X):
            summary = get_split_info(X, i[0], j, "train", summary)
            summary = get_split_info(X, i[1], j, "valid", summary)
            j = j + 1

        return pd.DataFrame(summary)


class KlusterFoldValidation:
    """
    KlusterFold cross validator

    - Create folds based on provided cluster method

    :param cluster_method: Clustering method with fit_predict attribute
    """

    def __init__(self, cluster_method=None):
        if not isinstance(cluster_method, Clusterer):
            raise ValueError(
                "The KlusterFoldValidation only works on cluster methods with .fit_predict."
            )

        self.cluster_method = cluster_method
        self.n_splits = None

    def split(self, X, y=None, groups=None):
        """
        Generator to iterate over the indices

        :param X: Array to split on
        :param y: Always ignored, exists for compatibility
        :param groups: Always ignored, exists for compatibility
        """

        X = check_array(X)

        if not self._method_is_fitted(X):
            self.cluster_method.fit(X)
        clusters = self.cluster_method.predict(X)

        self.n_splits = len(np.unique(clusters))

        if self.n_splits < 2:
            raise ValueError(
                f"Clustering method resulted in {self.n_splits} cluster, too few for fold validation"
            )

        for label in np.unique(clusters):
            yield (
                np.where(clusters != label)[0],
                np.where(clusters == label)[0],
            )

    def _method_is_fitted(self, X):
        """
        :param X: Array to use if the method is fitted
        :return: True if fitted, else False
        """
        try:
            self.cluster_method.predict(X[0:1, :])
            return True
        except NotFittedError:
            return False


class GroupTimeSeriesSplit(_BaseKFold):
    """
    Sliding window time series split

    Create n_splits folds with an as equally possible size through a smart variant of a brute
    force search. Groups parameter in .split() should be filled with the time groups (e.g. years)

    If n_splits is 3 ("*" = train, "x" = test)::

    |-----------------------|
    | * * * x x x - - - - - |
    | - - - * * * x x x - - |
    | - - - - - - * * * x x |
    |-----------------------|

    :param n_splits: the amount of train-test combinations.
    :type n_splits: int
    """

    # table above inspired by sktime

    def __init__(self, n_splits):
        if not isinstance(n_splits, numbers.Integral):
            raise ValueError(
                "The number of folds must be of Integral type. "
                "%s of type %s was passed." % (n_splits, type(n_splits))
            )
        n_splits = int(n_splits)

        if n_splits <= 1:
            raise ValueError(
                "k-fold cross-validation requires at least one"
                " train/test split by setting n_splits=2 or more,"
                " got n_splits={0}.".format(n_splits)
            )

        self.n_splits = n_splits

    def summary(self):
        """
        Generates a pd.DataFrame which displays the groups splits
        and extra statistics about it.
        Can only be run after having applied the .split() method to
        the GroupTimeSeriesSplit instance.

        :return: a pd.DataFrame with info about where the splits were made
        :rtype: pd.DataFrame
        """
        try:
            return (
                self._grouped_df.sort_index()
                .assign(group=lambda df: df["group"].astype(int))
                .assign(
                    obs_per_group=lambda df: df.groupby("group")[
                        "observations"
                    ].transform("sum")
                )
                .assign(ideal_group_size=round(self._ideal_group_size))
                .assign(
                    diff_from_ideal_group_size=lambda df: df["obs_per_group"]
                    - df["ideal_group_size"]
                )
            )
        except AttributeError:
            raise AttributeError(
                ".summary() only works after having ran" " .split(X, y, groups)."
            )

    def split(self, X=None, y=None, groups=None):
        """Returns the train-test splits of all the folds

        :param X: array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features, defaults to None
        :type X: np.array, optional
        :param y: array-like of shape (n_samples,)
            The target variable for supervised learning problems, defaults to None
        :type y: np.array, optional
        :param groups: Group labels for the samples used while splitting the dataset
            into train/test set, defaults to None
        :type groups: np.array
        :return: the indices of the train and test splits
        :rtype: List[np.array]
        """
        if groups is None:
            raise ValueError("Groups cannot be None")
        X, y, groups = indexable(X, y, groups)
        n_groups = np.unique(groups).shape[0]
        if self.n_splits >= n_groups:
            raise ValueError(
                (
                    "n_splits({0}) must be less than the amount"
                    " of unique groups({1})."
                ).format(self.n_splits, n_groups)
            )
        return list(self._iter_test_indices(X, y, groups))

    def get_n_splits(self, X=None, y=None, groups=None):
        """Get the amount of splits

        :param X: Always ignored, exists for compatibality, defaults to None
        :type X: Object, optional
        :param y: Always ignored, exists for compatibality, defaults to None
        :type y: Object, optional
        :param groups: Always ignored, exists for compatibality, defaults to None
        :type groups: Object, optional
        :return: amount of n_splits
        :rtype: int
        """
        return self.n_splits

    def _check_for_long_estimated_runtime(self, groups):
        """
        Checks for combinations of n_splits and unique groups and raises UserWarning
        if runtime is expected to take over one minute

        :param groups: array of the groups
        :type groups: np.array
        """
        unique_groups = len(set(groups))
        warning = (
            "Finding the optimal split points"
            " with {0} unique groups and n_splits at {1}"
            " can take several minutes."
        ).format(unique_groups, self.n_splits)
        if self.n_splits == 4 and unique_groups > 250:
            warn(
                warning + " Consider to decrease n_splits to 3 or lower.",
                UserWarning,
            )

        elif self.n_splits == 5 and unique_groups > 100:
            warn(
                warning + " Consider to decrease n_splits to 4 or lower.",
                UserWarning,
            )

        elif self.n_splits > 5 and unique_groups > 30:
            warn(
                warning + " Consider to decrease n_splits to 5 or lower.",
                UserWarning,
            )

    def _iter_test_indices(self, X=None, y=None, groups=None):
        """
        Calculates the optimal division of groups into folds so that every fold is as
        equally large as possible

        :param X: Always ignored, exists for compatibility.
        :param y: Always ignored, exists for compatibility.
        :param groups: array with groups
        :type groups: np.array

        :yields: two numpy arrays -> train indices and test indices of the same fold
        """
        self._check_for_long_estimated_runtime(groups)
        (
            self._first_split_index,
            self._last_split_index,
        ) = self._calc_first_and_last_split_index(groups=groups)
        self._best_splits = self._get_split_indices()
        groups = self._regroup(groups)
        for i in range(self.n_splits):
            yield np.where(groups == i)[0], np.where(groups == i + 1)[0]

    def _calc_first_and_last_split_index(self, X=None, y=None, groups=None):
        """
        Calculates an approximate first and last split point to reduce the amount
        of options during a brute force search.

        :param X: Always ignored, exists for compatibility.
        :param y: Always ignored, exists for compatibility.
        :param groups: array with groups
        :type groups: np.array

        :return: approximate index of first and approx. index of last split
        :rtype: tuple
        """

        # get the counts (=amount of rows) for each group
        self._grouped_df = (
            pd.DataFrame(np.array(groups))
            .rename(columns={0: "index"})
            .groupby("index")
            .size()
            .sort_index()
            .to_frame()
            .rename(columns={0: "observations"})
        )

        # set the ideal group_size and reduce it to 90% to have some leverage
        self._ideal_group_size = np.sum(self._grouped_df["observations"]) / (
            self.n_splits + 1
        )
        init_ideal_group_size = self._ideal_group_size * 0.9

        # initialize the index of the first split, to reduce the amount of possible index split options
        first_split_index = (
            self._grouped_df.assign(cumsum_obs=lambda df: df["observations"].cumsum())
            .assign(group_id=lambda df: (df["cumsum_obs"] - 1) // init_ideal_group_size)
            .reset_index()
            .loc[lambda df: df["group_id"] != 0]
            .iloc[0]
            .name
        )
        # initialize the index of the last split point, to reduce the amount of possible index split options
        last_split_index = len(self._grouped_df) - (
            self._grouped_df.assign(
                observations=lambda df: df["observations"].values[::-1],
                cumsum_obs=lambda df: df["observations"].cumsum(),
            )
            .reset_index()
            .assign(group_id=lambda df: (df["cumsum_obs"] - 1) // init_ideal_group_size)
            .loc[lambda df: df["group_id"] != 0]
            .iloc[0]
            .name
            - 1
        )
        return first_split_index, last_split_index

    def _get_split_indices(self):
        """
        Calculates for each possible splits the total absolute different of the groups
        to the ideal group size and saves the split with the least absolute difference.

        :return: indices of the best split points
        :rtype: tuple
        """
        # set the index range to search possible splits for
        index_range = range(self._first_split_index, self._last_split_index)

        observations = self._grouped_df["observations"].tolist()

        # create generator with all the possible index splits
        # e.g. for [0, 1, 3, 5, 8] and self.n_splits = 2
        # [(1,2), (1,3), (1,4), (2,3), (2,4), (3,4)]
        # with for the first split:
        # group1 = [:1]
        # group2 = [1:2]
        # group3 = [2:]
        splits_generator_shifted = combinations(index_range, self.n_splits)

        # get the first iteration
        first_splits = next(splits_generator_shifted)

        # create a new generator that starts from the beginning again
        splits_generator = combinations(index_range, self.n_splits)

        # generate a list, with for every group the difference between them and the ideal group size
        # e.g.
        # ideal_group_size = 100
        # group_sizes = [10,20,270]
        # diff_from_ideal_list = [-90, -80, 170]
        diff_from_ideal_list = [
            sum(observations[: first_splits[0]]) - self._ideal_group_size
        ]
        for split in sliding_window(first_splits, window_size=2, step_size=1):
            try:
                diff_from_ideal_list += [
                    sum(observations[split[0] : split[1]]) - self._ideal_group_size
                ]
            except IndexError:
                diff_from_ideal_list += [
                    sum(observations[split[0] :]) - self._ideal_group_size
                ]

        # keep track of the minimum of the total difference from all groups to the ideal group size
        min_diff = sum([abs(diff) for diff in diff_from_ideal_list])
        best_splits = first_splits

        # loop through all possible split points and check whether a new split
        # has a less total difference from all groups to the ideal group size
        for prev_splits, new_splits in zip(splits_generator, splits_generator_shifted):
            diff_from_ideal_list = self._calc_new_diffs(
                observations, diff_from_ideal_list, prev_splits, new_splits
            )
            new_diff = sum([abs(diff) for diff in diff_from_ideal_list])

            # if with the new split the difference is less than the current most optimal, save the
            # new split
            if new_diff < min_diff:
                min_diff = new_diff
                best_splits = new_splits
        return best_splits

    @staticmethod
    def _calc_new_diffs(values, diff_list, prev_splits, new_splits):
        """Calculates the new group size differences compared to the optimal group size

        :param values: list of values
        :type values: list
        :param diff_list: list of values with for each index split its difference
            from the optimal group size
        :type diff_list: list
        :param prev_splits: the indices of the previous splits, excluding index 0 and the last index
        :type prev_splits: tuple
        :param new_splits: the indices of the new splits, excluding index 0 and the last index
        :type new_splits: tuple

        :return: updated diff_list
        :rtype: list
        """
        # calculate which indices have changed, e.g.:
        # new_index = (1,2,5)
        # prev_index = (1,2,4)
        # index_diffs = (0,0,1)
        index_diffs = [
            new_index - prev_index
            for prev_index, new_index in zip(prev_splits, new_splits)
        ]
        new_diff_list = diff_list.copy()

        # calculate the effects of the index change to the groups
        for index, diff in enumerate(index_diffs):
            if diff != 0:
                start_index, end_index = (
                    (prev_splits[index], new_splits[index])
                    if prev_splits[index] < new_splits[index]
                    else (new_splits[index], prev_splits[index])
                )

                # calculate the value change from one group to another
                value_change = sum(values[start_index:end_index])

                # if diff < 0 the previous group gains values, so change value_change to -value_change
                value_change = value_change if diff > 0 else -value_change

                # change the values of the current and next group
                new_diff_list[index] += value_change
                new_diff_list[index + 1] -= value_change

        return new_diff_list

    def _regroup(self, groups):
        """
        Specifies in which group every observation belongs

        :param groups: original groups in array
        :type: groups: np.array

        :return: indices for the train and test splits of each fold
        :rtype: np.array
        """
        df = self._grouped_df.copy().reset_index()
        # set each unique group to the right group_id to group them into folds
        df.loc[: self._best_splits[0], "group"] = 0
        for group_id, splits in enumerate(sliding_window(self._best_splits, 2, 1)):
            try:
                df.loc[splits[0] : splits[1], "group"] = group_id + 1
            except IndexError:
                df.loc[splits[0] :, "group"] = group_id + 1

        self._grouped_df = df
        # create a mapper to set every group to the right group_id
        mapper = dict(zip(df["index"], df["group"]))
        return np.vectorize(mapper.get)(groups)

    def _method_is_fitted(self, X):
        """
        :param X: Array to use if the method is fitted
        :return: True if fitted, else False
        """
        try:
            self.cluster_method.predict(X[0:1, :])
            return True
        except NotFittedError:
            return False



import pandas as pd
from pandas.testing import assert_frame_equal
from sklearn.pipeline import make_pipeline
import pytest
from sklego.preprocessing import ColumnSelector


@pytest.fixture()
def df():
    return pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6],
            "b": [10, 9, 8, 7, 6, 5],
            "c": ["a", "b", "a", "b", "c", "c"],
            "d": ["b", "a", "a", "b", "a", "b"],
            "e": [0, 1, 0, 1, 0, 1],
        }
    )


def test_select_two(df):
    result_df = ColumnSelector(["d", "e"]).fit_transform(df)
    expected_df = pd.DataFrame(
        {"d": ["b", "a", "a", "b", "a", "b"], "e": [0, 1, 0, 1, 0, 1]}
    )

    assert_frame_equal(result_df, expected_df)


def test_select_one(df):
    result_df = ColumnSelector(["e"]).fit_transform(df)
    expected_df = pd.DataFrame({"e": [0, 1, 0, 1, 0, 1]})

    assert_frame_equal(result_df, expected_df)


def test_select_all(df):
    result_df = ColumnSelector(["a", "b", "c", "d", "e"]).fit_transform(df)
    assert_frame_equal(result_df, df)


def test_select_none(df):
    with pytest.raises(ValueError):
        ColumnSelector([]).fit_transform(df)


def test_select_not_in_frame(df):
    with pytest.raises(KeyError):
        ColumnSelector(["f"]).fit_transform(df)


def test_select_one_in_pipeline(df):
    pipe = make_pipeline(ColumnSelector(["d"]))
    result_df = pipe.fit_transform(df)
    expected_df = pd.DataFrame({"d": ["b", "a", "a", "b", "a", "b"]})

    assert_frame_equal(result_df, expected_df)


def test_get_feature_names():
    df = pd.DataFrame({"a": [4, 5, 6], "b": ["4", "5", "6"]})
    transformer = ColumnSelector("a").fit(df)
    assert transformer.get_feature_names() == ["a"]



from inspect import signature
import logging

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin, clone, is_regressor, is_classifier
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array
from sklearn.exceptions import NotFittedError


class ZeroInflatedRegressor(BaseEstimator, RegressorMixin):
    """
    A meta regressor for zero-inflated datasets, i.e. the targets contain a lot of zeroes.

    `ZeroInflatedRegressor` consists of a classifier and a regressor.

        - The classifier's task is to find of if the target is zero or not.
        - The regressor's task is to output a (usually positive) prediction whenever the classifier indicates that the
        there should be a non-zero prediction.

    The regressor is only trained on examples where the target is non-zero, which makes it easier for it to focus.

    At prediction time, the classifier is first asked if the output should be zero. If yes, output zero.
    Otherwise, ask the regressor for its prediction and output it.

    Parameters
    ----------
    classifier : Any, scikit-learn classifier
        A classifier that answers the question "Should the output be zero?".

    regressor : Any, scikit-learn regressor
        A regressor for predicting the target. Its prediction is only used if `classifier` says that the output is non-zero.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
    >>> np.random.seed(0)
    >>> X = np.random.randn(10000, 4)
    >>> y = ((X[:, 0]>0) & (X[:, 1]>0)) * np.abs(X[:, 2] * X[:, 3]**2)
    >>> z = ZeroInflatedRegressor(
    ... classifier=ExtraTreesClassifier(random_state=0),
    ... regressor=ExtraTreesRegressor(random_state=0)
    ... )
    >>> z.fit(X, y)
    ZeroInflatedRegressor(classifier=ExtraTreesClassifier(random_state=0),
                          regressor=ExtraTreesRegressor(random_state=0))
    >>> z.predict(X)[:5]
    array([4.91483294, 0.        , 0.        , 0.04941909, 0.        ])
    """

    def __init__(self, classifier, regressor) -> None:
        """Initialize."""
        self.classifier = classifier
        self.regressor = regressor

    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The training data.

        y : np.ndarray, 1-dimensional
            The target values.

        sample_weight : Optional[np.array], default=None
            Individual weights for each sample.

        Returns
        -------
        ZeroInflatedRegressor
            Fitted regressor.

        Raises
        ------
        ValueError
            If `classifier` is not a classifier or `regressor` is not a regressor.
        """
        X, y = check_X_y(X, y)
        self._check_n_features(X, reset=True)
        if not is_classifier(self.classifier):
            raise ValueError(
                f"`classifier` has to be a classifier. Received instance of {type(self.classifier)} instead.")
        if not is_regressor(self.regressor):
            raise ValueError(f"`regressor` has to be a regressor. Received instance of {type(self.regressor)} instead.")

        try:
            check_is_fitted(self.classifier)
            self.classifier_ = self.classifier
        except NotFittedError:
            self.classifier_ = clone(self.classifier)

            if "sample_weight" in signature(self.classifier_.fit).parameters:
                self.classifier_.fit(X, y != 0, sample_weight=sample_weight)
            else:
                logging.warning("Classifier ignores sample_weight.")
                self.classifier_.fit(X, y != 0)

        non_zero_indices = np.where(self.classifier_.predict(X) == 1)[0]

        if non_zero_indices.size > 0:
            try:
                check_is_fitted(self.regressor)
                self.regressor_ = self.regressor
            except NotFittedError:
                self.regressor_ = clone(self.regressor)

                if "sample_weight" in signature(self.regressor_.fit).parameters:
                    self.regressor_.fit(
                        X[non_zero_indices],
                        y[non_zero_indices],
                        sample_weight=sample_weight[non_zero_indices] if sample_weight is not None else None
                    )
                else:
                    logging.warning("Regressor ignores sample_weight.")
                    self.regressor_.fit(
                        X[non_zero_indices],
                        y[non_zero_indices],
                    )
        else:
            raise ValueError(
                """The predicted training labels are all zero, making the regressor obsolete. Change the classifier
                or use a plain regressor instead."""
            )

        return self

    def predict(self, X):
        """
        Get predictions.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Samples to get predictions of.

        Returns
        -------
        y : np.ndarray, shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)

        output = np.zeros(len(X))
        non_zero_indices = np.where(self.classifier_.predict(X))[0]

        if non_zero_indices.size > 0:
            output[non_zero_indices] = self.regressor_.predict(X[non_zero_indices])

        return output



import numpy as np
import pytest
from sklearn.model_selection import train_test_split

from sklego.common import flatten
from sklego.preprocessing import RandomAdder


from tests.conftest import select_tests, transformer_checks, nonmeta_checks, general_checks


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, transformer_checks, nonmeta_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_methods_subset_invariance",
            "check_transformer_data_not_an_array",
            "check_transformer_general",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
def test_estimator_checks(test_fn):
    adder = RandomAdder()
    test_fn(RandomAdder.__name__, adder)


def test_dtype_regression(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    assert RandomAdder().fit(X, y).transform(X).dtype == float


def test_dtype_classification(random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    assert RandomAdder().fit(X, y).transform(X).dtype == float


def test_only_transform_train(random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    X_train, X_test, y_train, y_test = train_test_split(X, y)

    random_adder = RandomAdder()
    random_adder.fit(X_train, y_train)

    assert np.all(random_adder.transform(X_train) != X_train)
    assert np.all(random_adder.transform(X_test) == X_test)



import numpy as np
import pytest

from sklego.common import flatten
from sklego.dummy import RandomRegressor
from tests.conftest import nonmeta_checks, regressor_checks, general_checks, select_tests


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, regressor_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_methods_subset_invariance",
            "check_regressors_train",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
def test_estimator_checks(test_fn):
    # Tests that are skipped:
    # 'check_methods_subset_invariance': Since we add noise, the method is not invariant on a subset
    # 'check_regressors_train': score is not always greater than 0.5 due to randomness
    regr_normal = RandomRegressor(strategy="normal")
    test_fn(RandomRegressor.__name__ + "_normal", regr_normal)

    regr_uniform = RandomRegressor(strategy="uniform")
    test_fn(RandomRegressor.__name__ + "_uniform", regr_uniform)


def test_values_uniform(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    mod = RandomRegressor(strategy="uniform")
    predictions = mod.fit(X, y).predict(X)
    assert (predictions >= y.min()).all()
    assert (predictions <= y.max()).all()
    assert mod.min_ == pytest.approx(y.min(), abs=0.0001)
    assert mod.max_ == pytest.approx(y.max(), abs=0.0001)


def test_values_normal(random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    mod = RandomRegressor(strategy="normal").fit(X, y)
    assert mod.mu_ == pytest.approx(np.mean(y), abs=0.001)
    assert mod.sigma_ == pytest.approx(np.std(y), abs=0.001)


def test_bad_values():
    np.random.seed(42)
    X = np.random.normal(0, 1, (10, 2))
    y = np.random.normal(0, 1, (10, 1))
    with pytest.raises(ValueError):
        RandomRegressor(strategy="foobar").fit(X, y)



import types

import pandas as pd
import warnings

import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import make_pipeline

from sklego.metrics import equal_opportunity_score
from sklego.preprocessing import ColumnSelector


def test_equal_opportunity_pandas():
    sensitive_classification_dataset = pd.DataFrame(
        {
            "x1": [1, 0, 1, 0, 1, 0, 1, 1],
            "x2": [0, 0, 0, 0, 0, 1, 1, 1],
            "y": [1, 1, 1, 0, 1, 0, 0, 1],
        }
    )

    X, y = (
        sensitive_classification_dataset.drop(columns="y"),
        sensitive_classification_dataset["y"],
    )

    mod_1 = types.SimpleNamespace()

    mod_1.predict = lambda X: np.array([1, 0, 1, 0, 1, 0, 1, 1])
    assert equal_opportunity_score(sensitive_column="x2")(mod_1, X, y) == 0.75

    mod_1.predict = lambda X: np.array([1, 0, 1, 0, 1, 0, 0, 1])
    assert equal_opportunity_score(sensitive_column="x2")(mod_1, X, y) == 0.75

    mod_1.predict = lambda X: np.array([1, 0, 1, 0, 1, 0, 0, 0])
    assert equal_opportunity_score(sensitive_column="x2")(mod_1, X, y) == 0


def test_p_percent_pandas_multiclass():
    sensitive_classification_dataset = pd.DataFrame(
        {
            "x1": [1, 0, 1, 0, 1, 0, 1, 1],
            "x2": [0, 0, 0, 0, 0, 1, 1, 1],
            "y": [1, 1, 1, 0, 1, 0, 0, 2],
        }
    )

    X, y = (
        sensitive_classification_dataset.drop(columns="y"),
        sensitive_classification_dataset["y"],
    )

    mod_1 = types.SimpleNamespace()

    mod_1.predict = lambda X: np.array([2, 0, 1, 0, 1, 0, 1, 2])
    assert (
        equal_opportunity_score(sensitive_column="x2", positive_target=2)(
            mod_1, X, np.array([2, 0, 1, 0, 1, 0, 1, 2])
        )
        == 1
    )

    mod_1.predict = lambda X: np.array([1, 0, 1, 0, 1, 0, 0, 1])
    assert (
        equal_opportunity_score(sensitive_column="x2", positive_target=2)(mod_1, X, y)
        == 0
    )

    mod_1.predict = lambda X: np.array([1, 0, 1, 0, 1, 0, 0, 0])
    assert (
        equal_opportunity_score(sensitive_column="x2", positive_target=2)(mod_1, X, y)
        == 0
    )


def test_p_percent_numpy(sensitive_classification_dataset):
    X, y = sensitive_classification_dataset
    X = X.values
    mod = LogisticRegression().fit(X, y)
    assert equal_opportunity_score(1)(mod, X, y) == 0


def test_warning_is_logged(sensitive_classification_dataset):
    X, y = sensitive_classification_dataset
    mod_fair = make_pipeline(ColumnSelector("x1"), LogisticRegression()).fit(X, y)
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        equal_opportunity_score("x2", positive_target=2)(mod_fair, X, y)
        assert issubclass(w[-1].category, RuntimeWarning)



import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import Lasso
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklego.model_selection import GroupTimeSeriesSplit

GROUPS_COL = "group"


@pytest.fixture
def init_data():
    X = np.random.randint(low=1, high=1000, size=17)
    y = np.random.randint(low=1, high=1000, size=17)
    groups = np.array(
        [2000] * 3
        + [2001]
        + [2002] * 2
        + [2003]
        + [2004] * 5
        + [2005] * 2
        + [2006] * 2
        + [2007]
    )
    return X, y, groups


@pytest.fixture(params=[(2), (3), (4), (5), (6), (7)])
def valid_n_splits(request):
    return request.param


@pytest.fixture
def valid_cv(init_data, valid_n_splits):
    X, y, groups = init_data
    cv = GroupTimeSeriesSplit(valid_n_splits)
    _ = cv.split(X, y, groups)
    return cv


@pytest.mark.parametrize("n_splits", [(1), ("a"), (2.2), ([5]), (-1)])
def test_invalid_n_splits(n_splits):
    with pytest.raises(ValueError):
        GroupTimeSeriesSplit(n_splits=n_splits)


@pytest.mark.parametrize("n_splits", [(10), (20), (30), (100)])
def test_too_large_n_splits(init_data, n_splits):
    X, y, groups = init_data
    cv = GroupTimeSeriesSplit(n_splits=n_splits)
    with pytest.raises(ValueError):
        cv.split(X, y, groups)


def test_error_summary_before_split():
    cv = GroupTimeSeriesSplit(2)
    with pytest.raises(AttributeError):
        cv.summary()


def test_summary_method(valid_cv):
    cv_summary = valid_cv.summary()
    assert (
        isinstance(cv_summary, pd.DataFrame) and not cv_summary.empty
    ), f"{type(cv_summary)}\n{cv_summary}"


def test_n_groups_equal_n_splits_plus_one(valid_cv):
    cv_summary = valid_cv.summary()
    assert (
        cv_summary.loc[:, GROUPS_COL].nunique() == valid_cv.get_n_splits() + 1
    ), f"{cv_summary}"


def test_split_points_chronological(valid_cv):
    splits = valid_cv._best_splits
    assert list(splits) == sorted(splits), f"{splits}"


@pytest.mark.parametrize(
    "n_splits, n_groups", [(4, 251), (5, 101), (6, 31), (7, 31)]
)
def test_user_warning(n_splits, n_groups):
    groups = list(range(n_groups))
    X = np.random.randint(1, 10000, size=len(groups))
    y = np.random.randint(1, 10000, size=len(groups))
    cv = GroupTimeSeriesSplit(n_splits)
    with pytest.warns(UserWarning):
        cv.split(X, y, groups)


@pytest.mark.parametrize(
    "X, y, groups",
    [
        (
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=11),
        ),
        (
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=11),
            np.random.randint(0, 10, size=10),
        ),
        (
            np.random.randint(0, 10, size=11),
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=10),
        ),
        (
            np.random.randint(0, 10, size=9),
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=11),
        ),
    ],
)
def test_different_shapes_x_y_groups(X, y, groups, valid_n_splits):
    cv = GroupTimeSeriesSplit(valid_n_splits)
    with pytest.raises(ValueError):
        cv.split(X, y, groups)


def test_works_with_only_groups(init_data, valid_n_splits):
    _, _, groups = init_data
    cv = GroupTimeSeriesSplit(valid_n_splits)
    assert cv.split(groups=groups)


def test_grouptimeseriessplit_with_gridsearch(valid_n_splits):
    X_train = np.random.randint(1, 1000, size=100)
    y_train = np.random.randint(1, 1000, size=100)
    groups_train = np.random.randint(0, 10, size=100)

    # X_train, y_train, groups_train = init_data
    X_train = X_train.reshape(-1, 1)
    cv_splits = GroupTimeSeriesSplit(valid_n_splits).split(
        X_train, y_train, groups_train
    )
    Lasso(random_state=0, tol=0.1, alpha=0.8).fit(
        X_train, y_train, groups_train
    )
    pipe = Pipeline([("reg", Lasso(random_state=0, tol=0.1))])
    alphas = [0.1, 0.5, 0.8]
    grid = GridSearchCV(pipe, {"reg__alpha": alphas}, cv=cv_splits)
    grid.fit(X_train, y_train)
    best_C = grid.best_estimator_.get_params()["reg__alpha"]
    assert best_C


def test_amount_of_n_splits(init_data, valid_n_splits):
    X, y, groups = init_data
    cv = GroupTimeSeriesSplit(valid_n_splits)
    assert len(cv.split(X, y, groups)) == valid_n_splits


def test_series_same_output_as_arrays(init_data, valid_n_splits):
    X, y, groups = init_data
    df = pd.DataFrame({"X": X, "y": y, "groups": groups})
    cv1 = GroupTimeSeriesSplit(valid_n_splits)
    cv2 = GroupTimeSeriesSplit(valid_n_splits)
    cv1.split(X, y, groups)
    summary1 = cv1.summary()
    cv2.split(df, df["y"], df["groups"])
    summary2 = cv2.summary()
    pd.testing.assert_frame_equal(summary1, summary2)



import datetime as dt
import inspect
from functools import partial, wraps

import numpy as np
import pandas as pd
from scipy.ndimage import shift

from sklego.common import as_list


def _get_shape_delta(old_shape, new_shape):
    diffs = [
        ("+" if new > old else "") + str(new - old)
        for new, old in zip(new_shape, old_shape)
    ]

    return f"delta=({', '.join(diffs)})"


def log_step(
    func=None,
    *,
    time_taken=True,
    shape=True,
    shape_delta=False,
    names=False,
    dtypes=False,
    print_fn=print,
    display_args=True,
    log_error=True,
):
    """
    Decorates a function that transforms a pandas dataframe to add automated logging statements

    :param func: callable, function to log, defaults to None
    :param time_taken: bool, log the time it took to run a function, defaults to True
    :param shape: bool, log the shape of the output result, defaults to True
    :param shape_delta: bool, log the difference in shape of input and output, defaults to False
    :param names: bool, log the names of the columns of the result, defaults to False
    :param dtypes: bool, log the dtypes of the results, defaults to False
    :param print_fn: callable, print function (e.g. print or logger.info), defaults to print
    :param print_args: bool, whether or not to print the arguments given to the function.
    :param log_error: bool, whether to add the Exception message to the log if the function fails, defaults to True.
    :returns: the result of the function

    :Example:
    >>> @log_step
    ... def remove_outliers(df, min_obs=5):
    ...     pass

    >>> @log_step(print_fn=logging.info, shape_delta=True)
    ... def remove_outliers(df, min_obs=5):
    ...     pass

    """

    if func is None:
        return partial(
            log_step,
            time_taken=time_taken,
            shape=shape,
            shape_delta=shape_delta,
            names=names,
            dtypes=dtypes,
            print_fn=print_fn,
            display_args=display_args,
            log_error=log_error,
        )

    names = False if dtypes else names

    @wraps(func)
    def wrapper(*args, **kwargs):
        if shape_delta:
            old_shape = args[0].shape
        tic = dt.datetime.now()

        optional_strings = []
        try:
            result = func(*args, **kwargs)
            optional_strings = [
                f"time={dt.datetime.now() - tic}" if time_taken else None,
                f"n_obs={result.shape[0]}, n_col={result.shape[1]}" if shape else None,
                _get_shape_delta(old_shape, result.shape) if shape_delta else None,
                f"names={result.columns.to_list()}" if names else None,
                f"dtypes={result.dtypes.to_dict()}" if dtypes else None,
            ]
            return result
        except Exception as exc:
            optional_strings = [
                f"time={dt.datetime.now() - tic}" if time_taken else None,
                "FAILED" + (f" with error: {exc}" if log_error else ""),
            ]
            raise
        finally:
            combined = " ".join([s for s in optional_strings if s])

            if display_args:
                func_args = inspect.signature(func).bind(*args, **kwargs).arguments
                func_args_str = "".join(
                    ", {} = {!r}".format(*item) for item in list(func_args.items())[1:]
                )
                print_fn(
                    f"[{func.__name__}(df{func_args_str})] " + combined,
                )
            else:
                print_fn(
                    f"[{func.__name__}]" + combined,
                )

    return wrapper


def log_step_extra(
    *log_functions,
    print_fn=print,
    **log_func_kwargs,
):
    """
    Decorates a function that transforms a pandas dataframe to add automated logging statements

    :param *log_functions: callable(s), functions that take the output of the decorated function and turn it into a log.
                                        Note that the output of each log_function is casted to string using `str()`
    :param print_fn: callable, print function (e.g. print or logger.info), defaults to print
    :param **log_func_kwargs: keyword arguments to be passed to log_functions
    :returns: the result of the function

    :Example:
    >>> @log_step_extra(lambda d: d["some_column"].value_counts())
    ... def remove_outliers(df, min_obs=5):
    ...     pass

    """
    if not log_functions:
        raise ValueError("Supply at least one log_function for log_step_extra")

    def _log_step_extra(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)

            func_args = inspect.signature(func).bind(*args, **kwargs).arguments
            func_args_str = "".join(
                ", {} = {!r}".format(*item) for item in list(func_args.items())[1:]
            )

            try:
                extra_logs = " ".join(
                    [str(log_f(result, **log_func_kwargs)) for log_f in log_functions]
                )
            except TypeError:
                raise ValueError(
                    f"All log functions should be callable, got {[type(log_f) for log_f in log_functions]}"
                )

            print_fn(
                f"[{func.__name__}(df{func_args_str})] " + extra_logs,
            )

            return result

        return wrapper

    return _log_step_extra


def add_lags(X, cols, lags, drop_na=True):
    """
    Appends lag column(s).

    :param X: array-like, shape=(n_columns, n_samples,) training data.
    :param cols: column name(s) or index (indices).
    :param lags: the amount of lag for each col.
    :param drop_na: remove rows that contain NA values.
    :returns: ``pd.DataFrame | np.ndarray`` with only the selected cols.

    :Example:

    >>> import pandas as pd
    >>> df = pd.DataFrame([[1, 2, 3],
    ...                    [4, 5, 6],
    ...                    [7, 8, 9]],
    ...                    columns=['a', 'b', 'c'],
    ...                    index=[1, 2, 3])

    >>> add_lags(df, 'a', [1]) # doctest: +NORMALIZE_WHITESPACE
       a  b  c  a1
    1  1  2  3  4.0
    2  4  5  6  7.0

    >>> add_lags(df, ['a', 'b'], 2) # doctest: +NORMALIZE_WHITESPACE
       a  b  c  a2   b2
    1  1  2  3  7.0  8.0

    >>> import numpy as np
    >>> X = np.array([[1, 2, 3],
    ...               [4, 5, 6],
    ...               [7, 8, 9]])

    >>> add_lags(X, 0, [1])
    array([[1, 2, 3, 4],
           [4, 5, 6, 7]])

    >>> add_lags(X, 1, [-1, 1])
    array([[4, 5, 6, 2, 8]])
    """

    # A single lag will be put in a list
    lags = as_list(lags)

    # Now we can iterate over the list to determine
    # whether it is a list of integers
    if not all(isinstance(x, int) for x in lags):
        raise ValueError("lags must be a list of type: " + str(int))

    # The keys of the allowed_inputs dict contain the allowed
    # types, and the values contain the associated handlers
    allowed_inputs = {
        pd.core.frame.DataFrame: _add_lagged_pandas_columns,
        np.ndarray: _add_lagged_numpy_columns,
    }

    # Choose the correct handler based on the input class
    for allowed_input, handler in allowed_inputs.items():
        if isinstance(X, allowed_input):
            return handler(X, cols, lags, drop_na)

    # Otherwise, raise a ValueError
    allowed_input_names = list(allowed_inputs.keys())
    raise ValueError("X type should be one of:", allowed_input_names)


def _add_lagged_numpy_columns(X, cols, lags, drop_na):
    """
    Append a lag columns.

    :param df: the input ``np.ndarray``.
    :param cols: column index / indices.
    :param drop_na: remove rows that contain NA values.
    :returns: ``np.ndarray`` with the concatenated lagged cols.
    """

    cols = as_list(cols)

    if not all([isinstance(col, int) for col in cols]):
        raise ValueError("Matrix columns are indexed by integers")

    if not all([col < X.shape[1] for col in cols]):
        raise KeyError("The column does not exist")

    combos = (shift(X[:, col], -lag, cval=np.NaN) for col in cols for lag in lags)

    # In integer-based ndarrays, NaN values are represented as
    # -9223372036854775808, so we convert back and forth from
    # original to float and back to original dtype
    original_type = X.dtype
    X = np.asarray(X, dtype=float)
    answer = np.column_stack((X, *combos))

    # Remove rows that contain NA values when drop_na is truthy
    if drop_na:
        answer = answer[~np.isnan(answer).any(axis=1)]

    # Change dtype back to its original
    answer = np.asarray(answer, dtype=original_type)
    return answer


def _add_lagged_pandas_columns(df, cols, lags, drop_na):
    """
    Append a lag columns.

    :param df: the input ``pd.DataFrame``.
    :param cols: column name(s).
    :param drop_na: remove rows that contain NA values.
    :returns: ``pd.DataFrame`` with the concatenated lagged cols.
    """

    cols = as_list(cols)

    # Indexes are not supported as pandas column names may be
    # integers themselves, introducing unexpected behaviour
    if not all([col in df.columns.values for col in cols]):
        raise KeyError("The column does not exist")

    combos = (
        df[col].shift(-lag).rename(col + str(lag)) for col in cols for lag in lags
    )

    answer = pd.concat([df, *combos], axis=1)

    # Remove rows that contain NA values when drop_na is truthy
    if drop_na:
        answer = answer.dropna()

    return answer



from sklearn.base import BaseEstimator, TransformerMixin

from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted


class IdentityTransformer(BaseEstimator, TransformerMixin):
    """
    The identity transformer returns what it is fed. Does not apply anything useful.
    The reason for having it is because you can build more expressive pipelines.

    :type check_X: bool, optional, default=False
    :param check_X: Whether to validate X to be non-empty 2D array of finite values
                    and attempt to cast X to float.
                    If disabled, the model/pipeline is expected to handle e.g. missing,
                    non-numeric, or non-finite values.
    """
    def __init__(
        self,
        check_X: bool = False
    ):
        self.check_X = check_X

    def fit(self, X, y=None):
        """'Fits' the estimator."""
        if self.check_X:
            X = check_array(X, copy=True, estimator=self)
        self.shape_ = X.shape
        return self

    def transform(self, X):
        """'Applies' the estimator."""
        if self.check_X:
            X = check_array(X, copy=True, estimator=self)
        check_is_fitted(self, 'shape_')
        if X.shape[1] != self.shape_[1]:
            raise ValueError(f"Wrong shape is passed to transform. Trained on {self.shape_[1]} cols got {X.shape[1]}")
        return X



import numpy as np
import pandas as pd
import pytest

from sklego.common import flatten
from sklego.preprocessing import DictMapper


from tests.conftest import select_tests, transformer_checks, general_checks, nonmeta_checks


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, transformer_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_dtype_object",
            "check_sample_weights_list",
            "check_sample_weights_pandas_series"
        ]
    )
)
def test_estimator_checks(test_fn):
    test_fn(DictMapper.__name__, DictMapper(mapper={"foo": 1}, default=-1))


@pytest.fixture()
def mapper():
    return {"foo": 1, "bar": 2, "baz": 3}


@pytest.mark.parametrize(
    "input_array,expected_array",
    [
        (["foo", "bar", "baz"], [1, 2, 3]),
        (["foo", "bar", "bar"], [1, 2, 2]),
        (["foo", "bar", "monty"], [1, 2, -1]),
        (["foo", "bar", np.nan], [1, 2, -1]),
        ([["foo", "bar", "baz"], ["foo", "bar", "baz"]], [[1, 2, 3], [1, 2, 3]]),
    ],
)
def test_array(input_array, expected_array, mapper):
    X = np.array(input_array).reshape(-1, 1)
    expected = np.array(expected_array).reshape(-1, 1)
    result = DictMapper(mapper=mapper, default=-1).fit_transform(X)
    np.testing.assert_array_equal(result, expected)


def test_pandas(mapper):
    X = pd.DataFrame(["foo", "bar", "baz"], dtype=object)
    expected = np.array([1, 2, 3]).reshape(-1, 1)
    result = DictMapper(mapper=mapper, default=-1).fit_transform(X)
    np.testing.assert_array_equal(result, expected)


def test_no_mapper():
    mapper = {}
    X = pd.DataFrame(["foo", "bar", "baz"], dtype=object)
    expected = np.array([-1, -1, -1]).reshape(-1, 1)
    result = DictMapper(mapper=mapper, default=-1).fit_transform(X)
    np.testing.assert_array_equal(result, expected)
